<?php
// $Id: transformations.data.widget.inc,v 1.3 2009/03/30 19:48:49 jpetso Exp $
/**
 * @file
 * Transformations -
 * A framework for generic schema-based data transformation pipelines.
 *
 * Copyright 2009 by Jakob Petsovits ("jpetso", http://drupal.org/user/56020)
 */

/**
 * Retrieve the most appropriate form element for a given operation input.
 * FALSE is returned if the given input key is not available for the operation.
 */
class TfDataWidget {
  /**
   * Retrieve the most appropriate form element for a given operation input.
   *
   * @param $operation
   *   The operation for whose input you want to find an appropriate widget.
   * @param $inputKey
   *   The input key inside the above operation that specifies the exact input.
   */
  function forInput(TfOperation $operation, $inputKey) {
    if (!in_array($inputKey, $operation->inputKeys())) {
      return FALSE;
    }
    $schema = $operation->inputProperty($inputKey, 'expectedSchema');
    $element = TfDataWidget::forSchema($schema, $operation->input($inputKey));

    $element['#title'] = check_plain($operation->inputProperty($inputKey, 'label'));
    $element['#required'] = $operation->inputProperty($inputKey, 'required');

    if (!empty($element['#default_value']) && !$operation->isInputSet($inputKey)) {
      $value = new TfDataWidgetValue($element);
      $value->setUseDefault(TRUE);
      $value->setValue($element['#default_value']);
      $element['#default_value'] = $value;
    }

    $description = check_plain($operation->inputProperty($inputKey, 'description'));
    if (!empty($description)) {
      $element['#description'] = $description;
    }
    return $element;
  }

  /**
   * Retrieve the most appropriate form element for a given schema.
   */
  function forSchema(TfSchemaElement $schema, TfDataWrapper $data = NULL) {
    $highest_priority = 0;
    $elements = array();

    foreach (module_implements('transformations_data_widgets') as $module) {
      $function = $module . '_transformations_data_widgets';
      $moduleElements = $function($schema, $data);

      foreach ($moduleElements as &$element) {
        if (!empty($element['#file']) && empty($element['#path'])) {
          $element['#path'] = drupal_get_path('module', $module);
        }
        $elements[] = $element;
      }
    }
    drupal_alter('transformations_data_widgets', $elements, $schema);

    $sorted_elements = array();

    foreach ($elements as $key => &$element) {
      $sorted_elements[$element['#priority']][] = $element;
      unset($element['#priority']);
    }
    // Sort the array, so that the highest priority widgets go first.
    krsort($sorted_elements);

    if (empty($sorted_elements)) {
      $element = array(
        '#type' => 'textarea',
        '#transformations_value_callback' => 'transformations_data_widget_php_code_value',
        '#title' => t('PHP code'),
        '#description' => t('<p>Enter PHP code to generate the desired value, and pass it back with a return statement. Example:</p><p><code>return $desired_value;</code></p>'),
      );
    }
    else {
      // All of the $highest_priority_widgets have the same priority, so choose
      // any random class out of that array - we just pick the first one.
      $highest_priority_elements = reset($sorted_elements);
      $element = reset($highest_priority_elements);
    }

    // Create the actual data widget that wraps the data-specific one and
    // provides additional logic for data management, validation and includes.
    $wrapper_element = array(
      '#type' => 'transformations_data_widget',
    );
    if (!empty($element['#file'])) {
      $wrapper_element['#transformations_include'] = array(
        'file' => $element['#file'],
        'path' => $element['#path'],
      );
      unset($element['#file']);
      unset($element['#path']);
    }
    if (isset($data)) {
      $wrapper_element['#default_value'] = $data;
    }

    // Migrate a couple more properties from the child widget to the wrapper.
    $wrapper_properties = array(
      '#transformations_value_callback',
      '#transformations_from_data_wrapper_callback',
    );
    foreach ($wrapper_properties as $key) {
      if (!empty($element[$key])) {
        $wrapper_element[$key] = $element[$key];
        unset($element[$key]);
      }
    }

    $wrapper_element['#transformations_child_element'] = $element;
    return $wrapper_element;
  }
}

/**
 * The return type of a 'transformations_data' widget.
 *
 * A class that stores an input widget's state for lazy evaluation, and only
 * retrieves the "real" value when being asked for it - as opposed to Form API
 * which is not able to distinguish between internal state (e.g. PHP code)
 * and result value (e.g. value resulting from that PHP code).
 */
class TfDataWidgetValue implements ArrayAccess {
  public $elementValue;
  public $valueCallback;
  public $fromWrapperCallback;
  public $includeFile;
  public $useDefault;

  public function __construct($element = NULL) {
    $this->valueCallback = isset($element['#transformations_value_callback'])
      ? $element['#transformations_value_callback']
      : NULL;
    $this->fromWrapperCallback = isset($element['#transformations_from_data_wrapper_callback'])
      ? $element['#transformations_from_data_wrapper_callback']
      : NULL;
    $this->includeFile = isset($element['#transformations_include'])
      ? $element['#transformations_include']
      : NULL;
    $this->useDefault = FALSE;
    $this->elementValue = NULL;
  }

  /**
   * Return TRUE if the default value should be used instead of the results of
   * the value() method, or FALSE otherwise.
   */
  public function useDefault() {
    return $this->useDefault;
  }
  public function setUseDefault($useDefault) {
    $this->useDefault = (boolean) $useDefault;
  }

  /**
   * Retrieve the value of the given element, run through the final
   * (Transformations-specific) value callback.
   */
  public function value() {
    if (!empty($this->useDefault)) {
      return NULL;
    }
    if (empty($this->valueCallback)) {
      return $this->elementValue;
    }
    $this->includeFile();
    $value = $this->valueCallback;
    return $value($this->elementValue);
  }

  /**
   * Set the actual TfDataWrapper value of the widget (intended for usage
   * from outside). The value is automatically converted to the widget's
   * internal data format by calling the "from data wrapper" callback.
   */
  public function setValue($value) {
    if (!is_object($value) || !($value instanceof TfDataWrapper)) {
      // Wrap the value in a TfDataWrapper if it's not already wrapped.
      // Slightly inefficient, but convenient for widgets as the callback
      // can rely on the data to be always passed in a consistent format.
      $value = TfDataWrapper::create($value);
    }

    if (empty($this->fromWrapperCallback)) {
      // Use TfDataWrapper::data() to extract the data. Widgets that don't
      // find this sufficient should provide their own callback.
      $value = isset($value) ? $value->data() : NULL;
    }
    else {
      // Use the "from data wrapper" callback if the widget has provided one.
      $this->includeFile();
      $fromWrapper = $this->fromWrapperCallback;
      $value = $fromWrapper($value);
    }

    $this->elementValue = $value;
  }

  private function includeFile() {
    if (!empty($this->includeFile)) {
      require_once './' . $this->includeFile['path'] . '/' . $this->includeFile['file'];
    }
  }

  /**
   * Retrieve the value of the given element.
   */
  public function childElementValue() {
    return $this->elementValue;
  }

  /**
   * Set the internal value of the widget (intended for usage from within
   * Form API and the widget).
   */
  public function setChildElementValue($value) {
    $this->elementValue = $value;
  }

  // Form API treats everything in '#tree' elements as array, so let's work
  // around this defiancy by implementing the ArrayAccess interface.
  public function offsetExists($key) {
    return in_array($key, array('widget', 'use_default'));
  }

  public function offsetGet($key) {
    if ($key == 'widget') {
      return $this->childElementValue();
    }
    elseif ($key == 'use_default') {
      return $this->useDefault();
    }
    throw new Exception('Invalid key.');
  }

  public function offsetSet($key, $value) {
    if ($key == 'widget') {
      $this->setChildElementValue($value);
    }
    elseif ($key == 'use_default') {
      $this->setUseDefault($value);
    }
    else {
      throw new Exception('Invalid key.');
    }
  }

  public function offsetUnset($key) {
    $this->offsetSet($key, ($key == 'use_default') ? FALSE : NULL);
  }
}

/**
 * Implementation of hook_transformations_data_widgets():
 * Provide a set of default widgets for the most common schemas.
 */
function transformations_transformations_data_widgets(TfSchemaElement $schema, $data = NULL) {
  static $path;

  if (!isset($path)) {
    $path = drupal_get_path('module', 'transformations') . '/core';
  }
  $widgets = array();

  if ($schema->hasFormat('php:type:string')) {
    $is_multiline = $schema->formatHint('string:multiline');
    $widgets[] = array(
      '#type' => !empty($is_multiline) ? 'textarea' : 'textfield',
      '#priority' => 1,
      '#required' => TRUE,
    );
  }

  $number_formats = array(
    'php:pseudo-type:number' => 'transformations_data_widget_decimal_validate',
    'php:type:integer' => 'transformations_data_widget_integer_validate',
    'php:type:float' => 'transformations_data_widget_decimal_validate',
    'php:type:double' => 'transformations_data_widget_decimal_validate',
  );
  foreach ($number_formats as $format => $validation_callback) {
    if ($schema->hasFormat($format)) {
      $widget = array(
        '#type' => 'textfield',
        '#priority' => 1,
        '#file' => 'transformations.data.widget.inc',
        '#path' => $path,
        '#element_validate' => $validation_callback,
      );
      if ($format == 'php:type:integer') {
        $widget['#transformations_value_callback'] = 'transformations_data_widget_integer_value';
      }
      elseif ($format == 'php:type:float') {
        $widget['#transformations_value_callback'] = 'transformations_data_widget_float_value';
      }
      elseif ($format == 'php:type:double') {
        $widget['#transformations_value_callback'] = 'transformations_data_widget_double_value';
      }
      $widgets[] = $widget;
    }
  }

  if ($schema->hasFormat('php:type:boolean')) {
    $widgets[] = array(
      '#type' => 'checkbox',
      '#priority' => 1,
      '#file' => 'transformations.data.widget.inc',
      '#path' => $path,
      '#transformations_value_callback' => 'transformations_data_widget_boolean_value',
    );
  }

  return $widgets;
}

/**
 * Element validation callback for numbers with decimal places allowed.
 */
function transformations_data_widget_decimal_validate(&$element, &$form_values) {
  if (!is_numeric($element['#value'])) {
    form_error($element, t('%element must contain a valid number.', array(
      '%element' => $element['#title'],
    )));
  }
}

/**
 * Widget validation callback for numbers with decimal places allowed.
 */
function transformations_data_widget_integer_validate(&$element, &$form_values) {
  if (((int) $element['#value']) != $element['#value']) {
    form_error($element, t('%element must contain a valid integer number.', array(
      '%element' => $element['#title'],
    )));
  }
}

function transformations_data_widget_php_code_value($element_value) {
  return eval($element_value);
}

function transformations_data_widget_integer_value($element_value) {
  return (int) $element_value;
}

function transformations_data_widget_float_value($element_value) {
  return (float) $element_value;
}

function transformations_data_widget_double_value($element_value) {
  return (double) $element_value;
}

function transformations_data_widget_boolean_value($element_value) {
  return (boolean) $element_value;
}


//
// Internal form element callbacks.
//

/**
 * Form value callback for the widget element.
 * (Redirected from transformations.module, hence the leading underscore.)
 */
function _transformations_data_widget_value($element, $edit = FALSE) {
  $value = new TfDataWidgetValue($element);

  // When $_POST is empty, the function is called with a single argument.
  if (func_num_args() == 1) {
    if (isset($element['#default_value'])) {
      $defaultValue = $element['#default_value'];

      if (is_object($defaultValue) && $defaultValue instanceof TfDataWidgetValue) {
        return $defaultValue;
      }
      $value->setUseDefault(FALSE);
      $value->setValue($defaultValue);
      return $value;
    }
    // No default value given -> default to the default value (if the widget is
    // not required, otherwise this setting is ignored anyways).
    $value->setUseDefault(TRUE);
    return $value;
  }

  // The form has been posted, so return the actual (lazy-wrapped) value.
  $value->setValue($edit['widget']);
  $value->setUseDefault(isset($edit['use_default']) ? $edit['use_default'] : FALSE);
  return $value;
}

/**
 * Process callback for the widget element.
 */
function transformations_data_widget_process($element, $edit, &$form_state, $complete_form) {
  $element['#tree'] = TRUE;

  // Due to the value callback, the value here is a TfDataWidgetValue object.
  $widgetValue = $element['#value'];

  $element['widget'] = $element['#transformations_child_element'];
  $element['widget']['#default_value'] = $widgetValue->childElementValue();

  if (isset($element['#title'])) {
    $element['widget']['#title'] = $element['#title'];
    unset($element['#title']);
  }

  if (isset($element['#description'])) {
    if (!isset($element['widget']['#description'])) {
      $element['widget']['#description'] = '';
    }
    $element['#description'] .= ' <p>' . check_plain($element['#description']) . '</p>';
    unset($element['#description']);
  }

  // Wrap the original '#element_validate' in our own wrapper that only
  // validates if the default value is not used.
  if (!empty($element['widget']['#element_validate'])) {
    $element['#element_validate'][] = 'transformations_data_widget_validate';
    $element['#transformations_validate'] = $element['widget']['#element_validate'];
    unset($element['widget']['#element_validate']);
  }

  if ($element['#required']) {
    $element['use_default'] = array(
      '#type' => 'value',
      '#value' => FALSE,
    );
  }
  else {
    $element['use_default'] = array(
      '#type' => 'checkbox',
      '#default_value' => $widgetValue->useDefault(),
      '#title' => t('Use the default value for this input.'),
    );
  }

  // Make the "required" star appear for the child widget even though it's not
  // formally required.
  $element['#transformations_child_required'] = $element['#required'];
  $element['#required'] = FALSE;
  $element['#pre_render'] = array('transformations_data_widget_pre_render');

  return $element;
}

/**
 * Validation callback for the widget element.
 */
function transformations_data_widget_validate(&$element, &$form_state) {
  if ($element['use_default']['#value']) {
    return; // the actual value doesn't matter, so skip widget validation
  }
  if (!empty($element['#transformations_include'])) {
    $include = $element['#transformations_include'];
    require_once './' . $include['path'] . '/' . $include['file'];
  }
  $validate = $element['#transformations_validate'];
  $validate($element['widget'], $form_state);
}

/**
 * Pre-render callback: Make the child element look like it's required.
 * (Which is the case actually, just doesn't go through '#required'.)
 */
function transformations_data_widget_pre_render($element) {
  $element['widget']['#required'] = $element['#transformations_child_required'];
  return $element;
}

/**
 * Theme function for the wrapper widget handling operation input values.
 */
function theme_transformations_data_widget($element) {
  return theme('form_element', $element, $element['#children']);
}
