<?php
// $Id: transformations.operation.inc,v 1.3 2009/03/28 12:24:32 jpetso Exp $
/**
 * @file
 * Transformations -
 * A framework for generic schema-based data transformation pipelines.
 *
 * Copyright 2009 by Jakob Petsovits ("jpetso", http://drupal.org/user/56020)
 */


/**
 * An instance of the TfOutput class is returned by TfOperation::output(),
 * and is used to gather all outputs of an operation in a nicely accessible
 * and safe enclosure.
 */
class TfOutput implements IteratorAggregate {
  private $outputKeys;
  private $operation;
  private $outputs;
  private $errorMessage;

  final public function __construct(TfOperation $operation) {
    $this->outputKeys = $operation->outputKeys();
    $this->operation = $operation;
    $this->outputs = array();
    $this->errorMessage = NULL;
  }

  /**
   * Return TRUE if the operation was successful and all outputs have been set,
   * or FALSE otherwise.
   */
  final public function isValid() {
    return isset($this->outputs);
  }

  /**
   * Make sure the operation is valid (@see TfOutput::isValid()), and throw
   * an exception if that is not the case.
   */
  private function assertValid() {
    if (!$this->isValid()) {
      $message = 'Operation ' . $this->operation->uiName() .
        ' was unsuccessful, no values can be retrieved with TfOutput::at().';
      if (!empty($this->errorMessage)) {
        $message .= ' ' . $this->errorMessage;
      }
      throw new Exception($message);
    }
  }

  /**
   * To be called by the operation in order to assign data to its outputs.
   * Must not be called after the operation has been executed.
   */
  final public function set($outputKey, $data) {
    if (!isset($this->outputKeys)) {
      throw new Exception('TfOutput::set() must not be called after the operation has been executed.');
    }
    if (!in_array($outputKey, $this->outputKeys)) {
      throw new Exception('Output key "' . $outputKey .
        '" was not specified by the output schema.');
    }
    if (!is_object($data) || !($data instanceof TfDataWrapper)) {
      $data = TfDataWrapper::create($data);
    }
    $this->outputs[$outputKey] = $data;
  }

  /**
   * Define the operation as failed, and specify a error message that explains
   * what went wrong. The message should not be run through output filters
   * like check_plain(), as the user of the class is expected to do that.
   */
  final public function setErrorMessage($errorMessage) {
    $this->operationDone(empty($errorMessage) ? t('(unknown error)') : $errorMessage);
  }

  /**
   * Retrieve the error message in case the operation has failed (which is the
   * case if the isValid() method returns FALSE). The message has not been run
   * through output filters like check_plain(), so the caller is expected to
   * care for proper sanitization of the error message. For a successfully
   * executed operation, no error message is set and this method returns FALSE.
   */
  final public function errorMessage() {
    return $this->isValid() ? FALSE : $this->errorMessage;
  }

  /**
   * Retrieve the keys for all output values. These keys can be passed to the
   * at() method to retrieve the corresponding output values.
   */
  final public function keys() {
    return array_keys($this->outputs);
  }

  /**
   * Retrieve the output value for the given @p $outputKey. Must not be called
   * if the isValid() method returns FALSE.
   */
  final public function at($outputKey) {
    $this->assertValid();

    if (!isset($this->outputs[$outputKey])) {
      throw new Exception('Data was not set for output key "' . $outputKey . '".');
    }
    return $this->outputs[$outputKey];
  }

  /**
   * Implementation of IteratorAggregate::getIterator().
   */
  final public function getIterator() {
    $this->assertValid();
    return new ArrayIterator($this->outputs);
  }

  /**
   * Finish up the creation of output values by assigning the expected schema
   * to each output value, and checking whether all expected outputs have
   * actually been set. Called by TfOperation::output() after the operation has
   * been executed, or by TfOutput::setErrorMessage() .
   */
  final public function operationDone($errorMessage = NULL) {
    if (isset($this->errorMessage)) {
      return; // This function has already been called before.
    }
    if (isset($errorMessage)) {
      // Set values so that isValid() == FALSE.
      unset($this->outputKeys);
      unset($this->outputs);
      $this->errorMessage = $errorMessage;
      return;
    }
    $outputs = array(); // for putting the outputs into correct order

    foreach ($this->outputKeys as $outputKey) {
      if (!isset($this->outputs[$outputKey])) {
        // Set values so that isValid() == FALSE.
        unset($this->outputKeys);
        unset($this->outputs);
        $this->errorMessage = t('Output key @key was not set by the operation.',
                                array('@key' => $outputKey));
        return;
      }

      // Output has been set, add it to the result array.
      if (is_object($this->outputs[$outputKey]) && ($this->outputs[$outputKey] instanceof TfDataWrapper)) {
        $outputs[$outputKey] = $this->outputs[$outputKey];
      }
      else {
        $outputs[$outputKey] = TfDataWrapper::create($this->outputs[$outputKey]);
      }
    }
    $this->outputs = $outputs;

    // Successful! Unset the stuff that we won't need anymore.
    unset($this->operation);
    unset($this->outputKeys);
  }
}

/**
 * Classes can implement this interface in order to get notified about changes
 * in the output schema of an operation (which is most probably triggered by a
 * change of the input schema). In order to get notified, the object has to be
 * registered with the operation by calling
 * TfOperation::registerSchemaOutputListener().
 */
interface TfOutputSchemaListener {
  /**
   * Called by a TfOperation if its output schema changes.
   *
   * @param $changedOperation
   *   The operation that changed part or all of its output schema.
   * @param $outputKey
   *   Unset if the whole schema changed, or the output key if just the schema
   *   of a single output element was affected by the change.
   */
  public function outputSchemaChanged(TfOperation $changedOperation, $outputKey = NULL);
}


/**
 * Base class for any data operation, with an arbitrary number of other
 * operations assigned for input retrieval and a single TfDataWrapper object
 * as output value.
 */
abstract class TfOperation {
  private $input;
  private $inputInfo;
  private $outputCache;
  private $outputInfo;
  private $outputSchemaListeners;

  //
  // Static methods.
  //

  /**
   * Retrieve a new instance of the given operation (defined by an
   * @p $operationSpec array that consists of the 'name' (class name) and
   * 'typeOptions' elements). Alternatively, you may also pass the class name
   * and type options as separate arguments.
   */
  public static function load($operationSpec, $typeOptions = NULL) {
    if (is_string($operationSpec)) {
      $operationSpec = array('name' => $operationSpec);
      if (isset($typeOptions)) {
        $operationSpec['typeOptions'] = $typeOptions;
      }
    }
    $className = $operationSpec['name'];
    $typeOptions = empty($operationSpec['typeOptions'])
                    ? array()
                    : $operationSpec['typeOptions'];

    // Ok, now that the specification is sorted out, load the operation.
    $operations = transformations_operation_info();
    if (!isset($operations[$className]) || !empty($operations[$className]['abstract'])) {
      return FALSE;
    }
    _transformations_include_extension($className, $operations);
    return new $className($typeOptions);
  }

  /**
   * Remove the given @p $prefix from each array key of the @p $typeOptions,
   * and unset any array elements that don't start with that prefix.
   *
   * @return
   *   The @p $typeOptions array after prefix removal.
   */
  public static function typeOptionsRemovePrefix($typeOptions, $prefix) {
    $unprefixedTypeOptions = array();
    foreach ($typeOptions as $key => $value) {
      if (strpos($key, $prefix) !== FALSE) {
        $originalKey = substr($key, strlen($prefix));
        $unprefixedTypeOptions[$originalKey] = $value;
      }
    }
    return $unprefixedTypeOptions;
  }


  //
  // Instance methods.
  //

  /**
   * Create and initialize a new instance of this TfOperation (sub)class.
   *
   * @see TfOperation::initialize()
   */
  final public function __construct(Array $typeOptions = array()) {
    $this->initialize($typeOptions);
  }

  /**
   * Return an array of type options that will allow this class to be recreated
   * with the same characteristics and input schema. This will be an empty
   * array for most operations, but some (e.g. operations taking other
   * operations as parameter) will require and provide such options.
   *
   * The set of allowed keys and values for this array is defined by the
   * hook_transformations_operation_info() implementation that registers the
   * operation, only strings or integers are allowed as type option values.
   */
  public function typeOptions() {
    return array();
  }

  /**
   * How the operation prefers to be called in the user interface.
   * Note that this name is not run through check_plain() or similar output
   * sanitization functions, the caller is expected to take care of that.
   *
   * Note that the name returned by this method is not necessarily used,
   * it's just a hint to the API user. By default, the 'label' property is
   * returned, which was (hopefully) specified in the operation-specific
   * implementation of [module]_operation_[class]().
   */
  public function uiName() {
    $info = transformations_operation_info(get_class($this));
    return ($info && isset($info['label'])) ? $info['label'] : get_class($this);
  }

  /**
   * Initialize this object. This is called from within the constructor but
   * can also be called to reset the object to its initial state.
   *
   * @param $typeOptions
   *   A structured array with options that influence the fundamental
   *   characteristics of the operation and affect the input schema.
   *   For example, a pipeline wrapper operation could take the pipeline id
   *   as type option, and the passed array might look like
   *   <code>array('pipeline' => $id)</code>. Invalid type options might be
   *   rejected by the operation with an exception being thrown.
   *
   *   TfOperation subclasses should only make use of type options if their
   *   input schema depends on these values - each possible combination of
   *   type options is regarded as separate operation, which quickly increases
   *   the effort that is required to propose matching operations. If it is
   *   possible for the TfOperation subclass to take options as regular
   *   operation inputs then by all means, do it that way.
   */
  public function initialize(Array $typeOptions = array()) {
    // Our own initializations.
    $this->input = array();
    $this->inputInfo = NULL;
    $this->outputSchemaListeners = array();
    $this->setOutputDirty();
    $this->setOutputSchemaDirty();
  }


  /**
   * @defgroup connection-info Input/output info
   * @{
   * Methods for retrieving information about expected input and output slots
   * and how the corresponding data is supposed to look like, as well as
   * methods for providing input and output schema information to the operation.
   */

  /**
   * Return an array of possible input keys for this operation., i.e. slots
   * that can be assigned input data. Among other things, these can be passed
   * to the setInput() method as key.
   *
   * Further information like the input's user-visible name, the expected
   * schema for the assigned data or whether the input is required or not
   * can be retrieved using the inputProperty() method.
   */
  final public function inputKeys() {
    $this->initInputKeys();
    return array_keys($this->inputInfo);
  }

  /**
   * Return an array of possible input keys for this operation.
   * Basically, this is the uncached version of inputKeys().
   */
  abstract protected function inputs();

  /**
   * Return an array of possible output keys for this operation., i.e. results
   * that can be retrieved from the TfOutput object that is returned by the
   * output() method.
   *
   * Further information like the output's user-visible name or the expected
   * schema for the output data can be retrieved using the outputProperty()
   * method.
   */
  final public function outputKeys() {
    $this->initOutputKeys();
    return array_keys($this->outputInfo);
  }

  /**
   * Return an array of possible input keys for this operation.
   * Basically, this is the uncached version of outputKeys().
   */
  abstract protected function outputs();


  /**
   * Return details on a specific operation input specified by @p $inputKey.
   * (If there is no such input key, an exception will be thrown. So make sure
   * that the inputKeys() method actually includes that key.)
   *
   * The @p $propertyKey parameter specifies which type of details should
   * be retrieved. Operations can provide any additional values, however the
   * following property keys are mandatory for any operation to provide
   * (or are provided by TfOperation itself) and thus will also be accessible:
   *
   * - 'label': A user-visible name for the input slot.
   * - 'required': If TRUE, the input must always be set with setInput().
   *      If FALSE, it can be skipped and a default value will be used instead.
   * - 'defaultValue': Retrieve the default input data for a given key (wrapped
   *      in a TfDataWrapper). NULL can be returned if the input is mandatory
   *      and there is no default value for this key.
   * - 'expectedSchema': The TfSchemaElement that the input value is expected
   *      to conform to. (If the input value does not conform to this schema,
   *      the operation behavior is undefined.)
   * - 'assignedSchema': If a more concrete TfSchemaElement has been set using
   *      the setInputSchema() method, that one will be returned. Otherwise,
   *      the same TfSchemaElement will be returned as for the 'expectedSchema'
   *      property key.
   *
   * A 'description' (string) property is not mandatory, but operations are
   * encouraged to provide one.
   *
   * If no property could be retrieved for the given @p $propertyKey,
   * NULL is returned.
   */
  final public function inputProperty($inputKey, $propertyKey) {
    $this->assertInputKeyExpected($inputKey);

    if (!array_key_exists($propertyKey, $this->inputInfo[$inputKey])) {
      // Special case: the only property handled directly by TfOperation.
      if ($propertyKey == 'assignedSchema') {
        return isset($this->inputInfo[$inputKey]['assignedSchema'])
          ? $this->inputInfo[$inputKey]['assignedSchema']
          : $this->inputInfo[$inputKey]['expectedSchema'];
      }

      // Retrieve the value from our subclass, and cache it in $this->inputInfo.
      $value = $this->inputInfo($inputKey, $propertyKey);

      if (!isset($value)) {
        if ($propertyKey == 'required') {
          $value = TRUE; // By default, inputs are required.
        }
        if ($propertyKey == 'label') {
          $value = $inputKey; // Should make for a bearable default label.
        }
        if ($propertyKey == 'expectedSchema') {
          throw new Exception('The ' . $this->uiName() .
            ' operation did not provide a schema for the input key ' .
            $inputKey . ' although that property is mandatory.');
        }
      }
      elseif ($propertyKey == 'defaultValue') { // && isset($value), from the elseif
        if (!(is_object($value) && $value instanceof TfDataWrapper)) {
          $value = TfDataWrapper::create($value);
        }
      }
      $this->inputInfo[$inputKey][$propertyKey] = $value;
    }
    return $this->inputInfo[$inputKey][$propertyKey];
  }

  /**
   * Return details on a specific operation input specified by @p $inputKey.
   * This function is only called by TfOperation::inputProperty(), and
   * essentially is the uncached version of that function. In order to avoid
   * bad surprises, the inputProperty() method makes sure that no other
   * @p $inputKey can be passed than those returned by inputs()/inputKeys().
   *
   * The @p $propertyKey parameter specifies which type of details should
   * be retrieved. Operations can provide any additional values, TfOperation
   * defines a couple of properties by itself:
   *
   * - 'label': A user-visible name for the input slot. Optional for the
   *      operation to provide - if NULL is returned by this function,
   *      inputProperty() will simply set the label to @p $inputKey.
   * - 'required': If TRUE, the input must always be set with setInput().
   *      If FALSE, it can be skipped and a default value will be used instead.
   *      Optional for the operation to provide - if NULL is returned by this
   *      function then inputProperty() will set the input to be required.
   * - 'defaultValue': Must be implemented by the subclass for mandatory inputs:
   *      Retrieve the default input data for a given key (either wrapped in a
   *      TfDataWrapper or not, TfOperation copes with both).
   * - 'expectedSchema': Must be implemented by the subclass:
   *      The TfSchemaElement that the input value is expected to conform to.
   *      (If the input value does not conform to this schema, the operation
   *      behavior is undefined.)
   *
   * A 'description' (string) property is not mandatory, but operations are
   * encouraged to provide one.
   *
   * If the operation doesn't define a value for the given @p $propertyKey,
   * NULL is returned.
   */
  abstract protected function inputInfo($inputKey, $propertyKey);

  /**
   * Return details on a specific operation output specified by @p $outputKey.
   * (If there is no such output key, an exception will be thrown. So make sure
   * that the outputKeys() method actually includes that key.)
   *
   * The @p $propertyKey parameter specifies which type of details should
   * be retrieved. Operations can provide any additional values, however the
   * following property keys are mandatory for any operation to provide
   * (or are provided by TfOperation itself) and thus will also be accessible:
   *
   * - 'label': A user-visible name for the output slot.
   * - 'expectedSchema': The TfSchemaElement that the output value is expected
   *      to conform to. The operation needs to make sure that this is the case.
   *
   * A 'description' (string) property is not mandatory, but operations are
   * encouraged to provide one.
   *
   * If no property could be retrieved for the given @p $propertyKey,
   * NULL is returned.
   */
  final public function outputProperty($outputKey, $propertyKey) {
    $this->assertOutputKeyExpected($outputKey);

    if (!array_key_exists($propertyKey, $this->outputInfo[$outputKey])) {
      // Retrieve the value from our subclass, and cache it in $this->outputInfo.
      $value = $this->outputInfo($outputKey, $propertyKey);

      if (!isset($value)) {
        if ($propertyKey == 'label') {
          $value = $outputKey; // Should make for a bearable default label.
        }
        if ($propertyKey == 'expectedSchema') {
          throw new Exception('The ' . $this->uiName() .
            ' operation did not provide a schema for the output key ' .
            $outputKey . ' although that property is mandatory.');
        }
      }
      $this->outputInfo[$outputKey][$propertyKey] = $value;
    }
    return $this->outputInfo[$outputKey][$propertyKey];
  }

  /**
   * Return details on a specific operation output specified by @p $outputKey.
   * This function is only called by TfOperation::outputProperty(), and
   * essentially is the uncached version of that function. In order to avoid
   * bad surprises, the outputProperty() method makes sure that no other
   * @p $outputKey can be passed than those returned by outputKeys().
   *
   * The @p $propertyKey parameter specifies which type of details should
   * be retrieved. Operations can provide any additional values, TfOperation
   * defines a couple of properties by itself:
   *
   * - 'label': A user-visible name for the input slot. Optional for the
   *      operation to provide - if NULL is returned by this function,
   *      outputProperty() will simply set the label to @p $outputKey.
   * - 'expectedSchema': Only mandatory property key for operations:
   *      The TfSchemaElement that the output value is expected to conform to.
   *      The operation needs to make sure that this is the case.
   *
   * A 'description' (string) property is not mandatory, but operations are
   * encouraged to provide one.
   *
   * If the operation doesn't define a value for the given @p $propertyKey,
   * NULL is returned.
   */
  abstract protected function outputInfo($inputKey, $propertyKey);


  /**
   * Specify the exact schema of the input data at input key @p $inputKey. This
   * schema must conform to the one given by
   * @code inputProperty($inputKey, 'expectedSchema'); @endcode
   * in other words, it must not be less detailed than the expected one.
   * If @p $schema is NULL, the more specific input schema for this element
   * will be reset to its less detailed originally expected state.
   *
   * @return
   *   TRUE if the input schema could be assigned, or FALSE if it didn't match
   *   the expected input schema for the given key.
   */
  public function setInputSchema($inputKey, TfSchemaElement $schema = NULL) {
    $this->assertInputKeyExpected($inputKey);
    ///TODO: Compare this schema with the expected one to make sure they match.

    if (isset($schema)) {
      $this->inputInfo[$inputKey]['assignedSchema'] = $schema;
      $this->inputSchemaChanged($inputKey);
    }
    elseif (isset($this->inputInfo[$inputKey]['assignedSchema'])) { // but not isset($schema)
      unset($this->inputInfo[$inputKey]['assignedSchema']);
      $this->inputSchemaChanged($inputKey);
    }
    // else {
    //   $this->inputInfo[$inputKey]['assignedSchema'] is unset before and
    //   after this function has been called, so nothing changes.
    // }
    return TRUE;
  }

  /**
   * Called when the input schema has been changed/refined using
   * setInputSchema(). Subclasses can override this function to avoid the
   * expected output schema to be reset (which is done by default).
   */
  protected function inputSchemaChanged($inputKey) {
    $this->setOutputSchemaDirty();
  }

  /**
   * Specify that the expected output schema has been refined with more
   * specificity, or put back to its original, less detailed state.
   * Preferably called when the input schema changes.
   *
   * @param $outputKey
   *   Unset if the whole schema changed, or the output key if just the schema
   *   of a single output element was affected by the change.
   */
  protected function setOutputSchemaDirty($outputKey = NULL) {
    $this->outputInfo = NULL;

    foreach ($this->outputSchemaListeners as $listener) {
      $listener->outputSchemaChanged($this, $outputKey);
    }
  }

  /**
   * Register a TfOutputSchemaListener object with this operation, so that it
   * will be notified when this operation's expected output schema changes.
   */
  final public function registerOutputSchemaListener(TfOutputSchemaListener $listener) {
    if (!empty($this->outputSchemaListeners)) {
      if (!in_array($listener, $this->outputSchemaListeners, TRUE /* strict */)) {
        $this->outputSchemaListeners[] = $listener;
      }
    }
  }

  /**
   * Unregister a TfOutputSchemaListener object from this operation.
   */
  final public function unregisterOutputSchemaListener(TfOutputSchemaListener $listener) {
    foreach ($this->outputSchemaListeners as $key => $existingListener) {
      if ($listener === $existingListener) {
        unset($this->outputSchemaListeners[$key]);
      }
    }
  }

  /** @} End of "defgroup connection-info". */


  /**
   * @defgroup input-data Input data
   * @{
   * Methods that deal with the actual input data of the operation, including
   * functionality to assign and retrieve that data.
   */

  /**
   * Assign a piece of data to this operation's input at @p $inputKey.
   *
   * If the @p $input parameter is NULL, the data for @p $inputKey will be
   * unset (and default values will be used, if any exist for the given key).
   *
   * If you pass an @p $input that is not a TfDataWrapper object, it will be
   * wrapped into a TfDataWrapper using the schema that is expected for that
   * input value. If @p $input does not conform to that schema, you're in
   * big trouble - make sure to never let this happen.
   */
  public function setInput($inputKey, $input = NULL) {
    $this->assertInputKeyExpected($inputKey);

    if (isset($input)) {
      if (!is_object($input) || !($input instanceof TfDataWrapper)) {
        $input = TfDataWrapper::create($input);
      }
      $this->input[$inputKey] = $input;
      $this->inputChanged($inputKey);
    }
    elseif (isset($this->input[$inputKey])) { // but not isset($input)
      unset($this->input[$inputKey]);
      $this->inputChanged($inputKey);
    }
    // else {
    //   $this->input is NULL before and after this function, no changes.
    // }
  }

  /**
   * Retrieve a piece of input data that was previously assigned to this
   * operation. Each piece of input data is represented as a TfDataWrapper.
   * If no associated input data exists, the default value for this input is
   * returned, or NULL if no default value is defined by the operation.
   */
  public function input($inputKey) {
    return $this->isInputSet($inputKey)
      ? $this->input[$inputKey]
      : $this->inputProperty($inputKey, 'defaultValue');
  }

  /**
   * Called when input data has been (re-)set using setInput(). Subclasses can
   * override this function to avoid the output cache to be reset (which is
   * done by default).
   */
  protected function inputChanged($inputKey) {
    $this->setOutputDirty();
  }

  /**
   * Specify that the operation needs to be (re-)executed.
   * Preferably called when input data changes.
   */
  protected function setOutputDirty() {
    $this->outputCache = NULL;
  }

  /**
   * Return TRUE if an input has been set via setInput(), or FALSE if it is
   * unset (even if there is a default value for that key).
   */
  public function isInputSet($inputKey) {
    return isset($this->input[$inputKey]);
  }

  /**
   * Return TRUE if all required input values have been set, or FALSE otherwise.
   */
  public function allRequiredInputsSet() {
    foreach ($this->inputKeys() as $inputKey) {
      if (!$this->isInputSet($inputKey) && $this->inputProperty($inputKey, 'required')) {
        return FALSE;
      }
    }
    return TRUE;
  }

  /** @} End of "defgroup input-data". */


  /**
   * @defgroup execution Operation execution
   * @{
   * Methods for actually executing the operation and influencing various
   * aspects of the execution process.
   */

  /**
   * In order to guarantee correct execution - such as, not being skipped for
   * performance reasons - operations with side effects need to override this
   * function and return TRUE instead. An operation is said to have side
   * effects if it takes any actions other than simply constructing the output
   * data, for example by writing data to a file or to the database.
   */
  public function hasSideEffects() {
    return FALSE;
  }

  /**
   * Override with "return TRUE;" to enable output caching.
   */
  protected function isOutputCachingEnabled() {
    return FALSE;
  }

  /**
   * Return the result of the operation as TfDataWrapper, given its current set
   * of input values. If the operation was unsuccessful, the isValid() method
   * of the resulting TfOutput object will return FALSE, otherwise you can
   * fetch the results via its at() method.
   */
  public function output() {
    if (!$this->allRequiredInputsSet()) {
      throw new Exception(
        'Not all required inputs are set for operation ' . $this->uiName() .
        '. (Inputs set: ' . implode(', ', array_keys($this->input)) . '.)'
      );
    }
    if ($this->isOutputCachingEnabled()) {
      if (isset($this->outputCache)) {
        return $this->outputCache;
      }
    }
    $output = new TfOutput($this);
    $this->execute($output);
    $output->operationDone();

    if ($this->isOutputCachingEnabled()) {
      $this->outputCache = $output;
    }
    return $output;
  }

  /**
   * Calculate the result of the operation (without concerns to caching) given
   * its current input values, and call $output->set() for each output value.
   * Call $output->setErrorMessage() if any errors occurred.
   */
  abstract protected function execute(TfOutput $output);

  /** @} End of "defgroup execution". */


  //
  // Private functions that are only of interest to TfOperation itself.
  //

  final private function initInputKeys() {
    if (!isset($this->inputInfo)) {
      $this->inputInfo = array();
      foreach ($this->inputs() as $key) {
        $this->inputInfo[$key] = array();
      }
    }
  }

  final private function initOutputKeys() {
    if (!isset($this->outputInfo)) {
      $this->outputInfo = array();
      foreach ($this->outputs() as $key) {
        $this->outputInfo[$key] = array();
      }
    }
  }

  /**
   * Make sure an input with the given @p $inputKey is among the possible input
   * slots, and throw an exception if that is not the case.
   */
  private function assertInputKeyExpected($inputKey) {
    $this->initInputKeys();
    if (!isset($this->inputInfo[$inputKey])) {
      throw new Exception('No input key "' . $inputKey .
        '" possible for operation ' . $this->uiName());
    }
  }

  /**
   * Make sure an output with the given @p $outputKey is among the possible
   * output slots, and throw an exception if that is not the case.
   */
  private function assertOutputKeyExpected($outputKey) {
    $this->initOutputKeys();
    if (!isset($this->outputInfo[$outputKey])) {
      throw new Exception('No output key "' . $outputKey .
        '" expected for operation ' . $this->uiName());
    }
  }
}
