<?php
// $Id: transformations.operation.tools.inc,v 1.1 2009/03/19 19:12:53 jpetso Exp $
/**
 * @file
 * Transformations -
 * A framework for generic schema-based data transformation pipelines.
 *
 * Copyright 2009 by Jakob Petsovits ("jpetso", http://drupal.org/user/56020)
 */


/**
 * This class contains a collection of helper functions that do not work
 * on actual TfOperation instances but on the static operation info that is
 * provided by the hooks in [module].transformations.inc and the plugin files.
 */
class TfOperationTools {
  const TypeOptionsDone = TRUE;
  const TypeOptionsMore = FALSE;

  /**
   * Get an array of all operations that can be instantiated, with the
   * class name as key and the operation label as value. This array can be
   * used as '#options' key of a 'select' type form element.
   *
   * @param $name
   *   If set, this function returns the operation label of the operation with
   *   this argument as class name, or FALSE if that operation is not available
   *   as an option.
   */
  public static function operationOptions($name = NULL) {
    $options = array();

    foreach (transformations_operation_info() as $className => $info) {
      if (!empty($info['abstract'])) {
        continue;
      }
      $options[$className] = $info['label'];
    }
    if (isset($name)) {
      return isset($options[$name]) ? $options[$name] : FALSE;
    }
    return $options;
  }

  /**
  * Transform a @p $typeOptionsInfo array (the one that is processed by
  * TfOperationTools::processTypeOptionsInfo() function) into a @p $typeOptions
  * array (the one that is returned by TfOperation::typeOptions() and can be
  * passed to TfOperation::load()).
  */
  public static function typeOptionsValuesFromInfo($typeOptionsInfo) {
    $typeOptions = array();
    foreach ($typeOptionsInfo as $key => $info) {
      if (!isset($typeOptionsInfo[$key]['value'])) {
        continue;
      }
      $typeOptions[$key] = $typeOptionsInfo[$key]['value'];
    }
    return $typeOptions;
  }

  /**
   * Transform a @p $typeOptions array (the one that is returned by
   * TfOperation::typeOptions() and can be passed to TfOperation::load()) into
   * into a @p $typeOptionsInfo array (the one that is processed by the
   * TfOperationTools::processTypeOptionsInfo() function).
   */
  public static function typeOptionsInfoFromValues($typeOptions) {
    $typeOptionsInfo = array();
    foreach ($typeOptions as $key => $value) {
      $typeOptionsInfo[$key] = array('value' => $value);
    }
    return $typeOptionsInfo;
  }

  /**
   * Execute the type options callback for a given operation. Might need to be
   * executed multiple times (with new options being assigned before this
   * function is called again) when one type option depends on another.
   *
   * This function also fills in a form element (or the title and description
   * properties of the form element, in case the callback has already provided
   * one) so after this function is called, you can use
   * $typeOptionsInfo[$key]['form_element'] to let the user choose the option.
   *
   * @param $className
   *   The class name of the operation whose type options should be retrieved.
   * @param $typeOptionsInfo
   *   An array of the form @code $typeOptionsInfo[$key][$property] @endcode
   *   that will potentially be amended with more type option keys and properties
   *   by this function and by the operation-specific type option callback.
   *   When called initially, this is expected to be an empty array.
   * @param $prefix
   *   A string prefix that is passed to the callback function which is expected
   *   to prepend it to all its type option keys. Makes it possible to include
   *   type options of "inner operations" without fear of key clashes.
   *
   * @return
   *   TfOperationTools::TypeOptionsDone if all type options have been assigned
   *   and no more need to be chosen, or TfOperationTools::TypeOptionsMore
   *   if the 'value' property still has to be assigned for at least one
   *   type option.
   */
  public static function processTypeOptionsInfo($className, &$typeOptionsInfo, $prefix = '') {
    $operations = transformations_operation_info();

    if (empty($operations[$className]['usesTypeOptions'])) {
      // No type options callback required, go home and be happy.
      return TfOperationTools::TypeOptionsDone;
    }
    _transformations_include_extension($className, $operations);
    $function = $operations[$className]['module'] . '_operation_' . $className . '_type_options';
    $function($typeOptionsInfo, $prefix);

    $status = TfOperationTools::TypeOptionsDone;

    foreach ($typeOptionsInfo as $key => $info) {
      if (!isset($info['options'])) { // called via recursion, leave alone for now
        continue;
      }
      if (count($info['options']) == 1) {
        $typeOptionsInfo[$key]['value'] = $info['value'] = reset(array_keys($info['options']));
      }
      if (!isset($info['value'])) {
        $status = TfOperationTools::TypeOptionsMore;
      }
    }
    return $status;
  }

  /**
   * After the @p $typeOptionsInfo array has been created or refined with
   * TfOperationTools::processTypeOptionsInfo() and that function has returned
   * TfOperationTools::TypeOptionsMore, create an input form that provides
   * selection elements for those type options that have not yet been assigned
   * a value.
   *
   * @param $typeOptionsInfo
   *   The array returned by TfOperationTools::processTypeOptionsInfo().
   *
   * @return
   *   A form element that uses the '#tree' property to get all type options
   *   into the same array on submitting the form. (Which means you can extract
   *   $typeOptions from $form_state['values'][$yourformkey].)
   */
  public static function typeOptionsForm($typeOptionsInfo) {
    $form = array(
      '#value' => '',
      '#tree' => TRUE,
    );

    foreach ($typeOptionsInfo as $key => $info) {
      if (isset($info['value'])) {
        $form[$key] = array(
          '#type' => 'value',
          '#value' => $info['value'],
        );
        $form[$key . '_transformations_description'] = array(
          '#type' => 'item',
          '#title' => $info['label'],
          '#value' => $info['options'][$info['value']],
        );
      }
      else {
        $form[$key] = array(
          '#type' => 'select',
          '#options' => $info['options'],
          '#default_value' => isset($info['default_value']) ? $info['default_value'] : NULL,
          '#title' => $info['label'],
          '#description' => isset($info['description']) ? $info['description'] : NULL,
          '#required' => TRUE,
        );
      }
    }
    return $form;
  }
}
