<?php
// $Id: transformations.schema.inc,v 1.2 2009/03/30 18:18:55 jpetso Exp $
/**
 * @file
 * Transformations -
 * A framework for generic schema-based data transformation pipelines.
 *
 * This file contains the classes that make up a schema.
 *
 * Copyright 2009 by Jakob Petsovits ("jpetso", http://drupal.org/user/56020)
 */


/**
 * Basics of schema structure in Transformations.
 *
 * Any concrete data element (TfData) is represented by a schema element
 * (TfSchemaElement), also referred to by "schema node" or just "schema".
 * Depending on the cardinality of collections that contain this element,
 * one schema element might correspond to multiple concrete data elements.
 *
 * A schema element essentially consists of:
 *   - a label (string), e.g. denoting the name of an XML element or the
 *       URL referring to an RDF statement,
 *   - a format (set of strings): descriptors denoting various characteristics
 *       that apply to the corresponding the data element, and
 *   - a child collection (TfSchemaCollection, optional), detailing the
 *       structure of the element if information about child elements is known.
 *
 * A schema element always refers to a single piece of data. If more than
 * a single data element shall be represented, you need a collection
 * (TfSchemaCollection) containing one or more schema elements.
 *
 * You should regard a TfSchemaCollection object more as a flat list than a
 * container element, it's a description of the order of the schema elements
 * ("members") that it holds. A TfSchemaCollection never corresponds to a
 * concrete data element, only the use of TfSchemaElement objects will
 * translate to an actual tree structure in the corresponding data.
 *
 * That way, it is possible to nest multiple collections without nesting the
 * data structure itself, e.g. a schema can describe the structure
 * "(student-properties | worker-properties) (first-name*) surname" with a
 * combination of different collections as members of a sequence collection:
 * a TfSchemaUnion (for student vs. worker), a TfSchemaCardinality with
 * cardinality [0, TfSchemaCardinality::Unbounded] (for the set of first names)
 * and a simple TfSchemaElement for the surname.
 *
 * If you think that this looks a lot like parser theory, you're right:
 * a generic schema is in fact nothing more than a context-free grammar, and
 * mapping data to schema elements is, as a consequence, a parsing problem.
 * (At least for the general case - there can be improvements for certain cases
 * where more assumptions can be taken about the structure of the data.)
 */


/**
 * Any part of a schema tree structure. Can be encapsulated in a proxy object
 * that forwards the given object's properties.
 */
interface TfSchemaPart {
  /**
   * In order to keep proxy generation both extensible (e.g. "create a tree of
   * TfQueryable* objects" vs. "create a tree of TfProxy* objects") and refrain
   * from hardcoding class names, every concrete schema part class has to
   * specify a base name for its proxies, e.g. "SchemaElement" or "SchemaSet".
   *
   * This base name will then be prefixed the name of the desired family of
   * proxy objects, which in the end results in a class name like
   * "TfProxySchemaElement" or "TfQueryableSchemaSet", which must derive
   * from TfSchemaProxy.
   */
  public function proxyBaseName();
}

/**
 * Abstract base class for all schema part proxies. The constructor takes two
 * parameters: the original TfSchemaPart object that should be encapsulated,
 * and the parent proxy object creating it. (The parent is of course optional,
 * as there must be an initial proxy without parent.)
 */
interface TfSchemaProxy {
  // Commented out, because the specialization of constructors is not really
  // compatible with polymorphism.
  //public function __construct(TfSchemaPart $original, TfSchemaProxy $parent = NULL);
}

/**
 * An interface, denoting that this object can assigned to the members() list
 * of a TfSchemaCollection. This interface is necessary because we don't want
 * TfSchemaPartialKnowledgeCollection objects (e.g. RDF data) to be members of
 * other collections as they make schema-based parsing pretty much impossible
 * (matching every element as valid). So that type of collection won't implement
 * this interface, and is only allowed as direct child of a TfSchemaElement.
 */
interface TfSchemaCollectionMember extends TfSchemaPart {
  /**
   * Return TRUE if at least one corresponding concrete data element must
   * exist, or FALSE otherwise.
   */
  public function isOptional();
}


//
// Elements.
//

/**
 * A schema element, representing a concrete piece of data.
 */
abstract class TfSchemaElement implements TfSchemaPart, TfSchemaCollectionMember {
  const ExactMatchRequired = 1;
  const NoMatchRequired = 0;

  /**
   * Dummy implementation, so that subclasses don't forget to
   * call parent::__construct(). Just in case we might need it.
   */
  public function __construct() {
  }

  /**
   * Implementation of TfSchemaPart::proxyBaseName().
   */
  public function proxyBaseName() {
    return "SchemaElement";
  }

  /**
   * Implementation of TfSchemaCollectionMember::isOptional().
   * A schema element always corresponds to a concrete data element,
   * and therefore is never optional.
   */
  final public function isOptional() {
    return FALSE;
  }

  /**
   * A list of formats that apply to this element.
   *
   * Such a format list normally consists of a PHP type
   * (e.g. 'php:type:string', 'php:type:array', 'php:type:object' or the
   * special type 'php:pseudo-type:number' for numeric values of any physical
   * data format) and optionally an arbitrary number of "meanings" that specify
   * a more exact role of the corresponding data (e.g. 'drupal:node:object'
   * or 'file:path').
   *
   * Formats like 'drupal:node:object' are a good way to specify data that
   * implements a fixed function but might look differently in terms of
   * object/array properties depending on the exact object. If it makes sense
   * to expose the members / child properties of such an object or array, exact
   * information on those can be put into a schema element's child collection
   * which is kind of an extended, more detailed version of the format list.
   *
   * It's also possible to have an empty list of formats, which means that no
   * information is known about the format of the corresponding data.
   *
   * Formats and member / child property information can be used to determine
   * whether or not different schemas are compatible to each other, and also
   * help with providing a suited data widget for data that corresponds to
   * this schema.
   *
   * Hierarchy levels for the scope of the format should be separated with
   * colons (e.g. 'php:class:TfSchemaElement') although that is merely
   * a convention.
   */
  abstract public function formats();

  /**
   * Check if a specific format applies to this element.
   *
   * @see TfSchemaElement::formats()
   */
  public function hasFormat($format) {
    return in_array($format, $this->formats());
  }

  /**
   * Retrieve a list of format hints.
   *
   * Unlike formats specified by the formats() method, format hints don't say
   * anything about the actual data format but only about the way that it
   * should be presented to the user, for example when entering data through
   * a data widget.
   *
   * The returned array has the format hint name (e.g. 'string:multiline')
   * as key, and the format hint value (e.g. TRUE) as corresponding element.
   *
   * Format hints should be optional and are named only by convention. If you
   * introduce format hints, try to cooperate with other module authors to
   * come up with common names and formats to improve the user experience.
   *
   * Also, make sure that format hint values are serializable so that schemas
   * can in turn be serialized themselves.
   */
  abstract public function formatHints();

  /**
   * Retrieve the value for a single format hint, or NULL if no format hint
   * with that name exists for this schema element.
   */
  public function formatHint($formatHintName) {
    $formatHints = $this->formatHints();
    return isset($formatHints[$formatHintName]) ? $formatHints[$formatHintName] : NULL;
  }

  /**
   * The label (name) of the element. Not necessarily similar to the label
   * of the concrete data, as a single schema element can correspond to
   * multiple data elements with different actual labels (e.g. the schema
   * element might have a label "row" while the concrete data has "1", "2",
   * or "row1", "row2", etc.).
   *
   * Whether or not (and if so, how) the schema element's label is reflected
   * by the concrete data element label(s) is checked by the isValidLabel()
   * method and specified by the labelRequirements() method.
   */
  abstract public function label();

  /**
   * Determine if the label @p $label a valid label for this schema element.
   *
   * @return
   *   TRUE if the given label matches this relation's label requirements,
   *   or FALSE otherwise.
   */
  public function isValidLabel($label) {
    $labelRequirements = $this->labelRequirements();

    if (is_string($labelRequirements)) {
      return (boolean) preg_match($labelRequirements, $label);
    }
    elseif ($labelRequirements == TfSchemaRelation::ExactMatchRequired) {
      return ($label == $this->label());
    }
    return TRUE;
  }

  /**
   * Determine if a this element's label must match any fixed string,
   * and if so, which string.
   *
   * @return
   *   Either of the following values:
   *
   *   - TfSchemaElement::ExactMatchRequired: The label of the concrete
   *       element must exactly match the one from this schema element.
   *   - TfSchemaElement::NoMatchRequired: The label of the concrete element
   *       does not matter, it can be anything.
   *   - A string value: Will be interpreted as PHP regular expression, the
   *       label of the concrete element must yield a preg_match() comparison
   *       with this string.
   */
  abstract public function labelRequirements();

  /**
   * Return NULL if this element doesn't possess any further structure (or the
   * structure of this element is not known), or a TfSchemaCollection that
   * describes the child elements of this element.
   */
  abstract public function childCollection();
}

/**
 * A read/write version of TfSchemaElement, for schema construction purposes.
 */
final class TfWritableSchemaElement extends TfSchemaElement {
  private $formats;
  private $formatHints;
  private $label;
  private $labelRequirements;
  private $childCollection;

  public function __construct($label, $formats = array()) {
    parent::__construct();

    $this->label = $label;
    $this->formats = array();
    $this->childCollection = NULL;

    if (!is_array($formats)) {
      $formats = func_get_args();
      array_shift($formats); // get rid of the $label argument in $formats
    }
    foreach ($formats as $format) {
      $this->setFormat($format);
    }
  }
  public function __clone() {
    if (isset($this->childCollection)) {
      $this->childCollection = clone $this->childCollection;
    }
  }

  // Formats.
  final public function formats() {
    return array_keys($this->formats);
  }
  final public function hasFormat($format) {
    return isset($this->formats[$format]);
  }
  final public function hasFormats($formats) {
    if (!is_array($formats)) {
      $formats = func_get_args();
    }
    foreach ($formats as $format) {
      if (!$this->hasFormat($format)) {
        return FALSE;
      }
    }
    return TRUE;
  }
  final public function setFormat($format) {
    $this->formats[$format] = TRUE;
  }
  final public function unsetFormat($format) {
    if (isset($this->formats[$format])) {
      unset($this->formats[$format]);
    }
  }

  // Format hints.
  final public function formatHints() {
    return $this->formatHints;
  }
  final public function formatHint($formatHintName) {
    return isset($this->formatHints[$formatHintName])
      ? $this->formatHints[$formatHintName]
      : NULL;
  }
  final public function setFormatHint($formatHintName, $value) {
    if (is_null($value)) {
      unset($this->formatHints[$formatHintName]);
    }
    else {
      $this->formatHints[$formatHintName] = $value;
    }
  }

  // Labels.
  final public function label() {
    return $this->label;
  }
  final public function setLabel($label) {
    $this->label = $label;
  }
  final public function labelRequirements() {
    return $this->labelRequirements;
  }
  final public function setLabelRequirements($labelRequirements) {
    $this->labelRequirements = $labelRequirements;
  }

  // Child collection.
  final public function childCollection() {
    return $this->childCollection;
  }
  /**
   * Set the child collection, and also return it (for more compact usage code).
   */
  final public function setChildCollection(TfSchemaCollection $childCollection) {
    $this->childCollection = $childCollection;
    return $this->childCollection;
  }
}

/**
 * Proxy class that encapsulates an actual TfSchemaElement and also generates
 * a proxy for its child collection on the fly.
 */
class TfProxySchemaElement extends TfSchemaElement implements TfSchemaProxy {
  protected $original;
  private $labelOverride;
  private $labelRequirementsOverride;
  private $childCollectionProxy;

  final public function __construct(TfSchemaElement $original, TfSchemaProxy $parent = NULL) {
    parent::__construct();
    $this->original = $original;
    $this->labelOverride = NULL;
    $this->labelRequirementsOverride = NULL;
    $this->childCollectionProxy = NULL;
    $this->setParent($parent);
  }
  protected function setParent(TfSchemaProxy $parent = NULL) {
    // Default implementation: do nothing.
  }

  public function __clone() {
    if (isset($this->childCollectionProxy)) {
      $this->childCollectionProxy = clone $this->childCollectionProxy;
    }
  }

  public function formats() {
    return $this->original->formats();
  }
  public function hasFormat($format) {
    return $this->original->hasFormat($format);
  }
  public function formatHints() {
    return $this->original->formatHints();
  }
  public function formatHint($formatHintName) {
    return $this->original->formatHint($formatHintName);
  }

  public function label() {
    if (isset($this->labelOverride)) {
      return $this->labelOverride;
    }
    return $this->original->label();
  }
  public function setLabel($label) {
    $this->labelOverride = $label;
    $this->labelRequirementsOverride = $this->original->labelRequirements();
  }
  public function isValidLabel($label) {
    if (isset($this->labelOverride)) {
      return parent::isValidLabel($label);
    }
    return $this->original->isValidLabel($label);
  }
  public function labelRequirements() {
    if (isset($this->labelOverride)) {
      return $this->labelRequirementsOverride;
    }
    return $this->original->labelRequirements();
  }

  public function childCollection() {
    if (!isset($this->childCollectionProxy)) {
      $originalChildCollection = $this->original->childCollection();

      if (!isset($originalChildCollection)) {
        $this->childCollectionProxy = NULL;
      }
      else {
        $baseName = $originalChildCollection->proxyBaseName();
        $className = $this->proxyClassName($baseName); // e.g. "TfProxySchemaSequence"
        $this->childCollectionProxy = new $className($originalChildCollection, $this);
      }
    }
    return $this->childCollectionProxy;
  }
  protected function proxyClassName($baseName) {
    return 'TfProxy' . $baseName;
  }
}


//
// Cardinality, a.k.a. optionality ("?") meets multiple occurrences ("+", "*").
//

/**
 * A collection representing repeated occurrences of elements relating to the
 * (single) member of this object, applying cardinality to the member. In terms
 * of parsing theory, this represents the Kleene star ("element*"),
 * optionality ("element?") and related concepts ("element+",
 * "element{lowerbound,upperbound}").
 */
abstract class TfSchemaCardinality implements TfSchemaCollectionMember {
  const Unbounded = -1;

  /**
   * Implementation of TfSchemaPart::proxyBaseName().
   */
  public function proxyBaseName() {
    return "SchemaCardinality";
  }

  /**
   * The minimum number of member instances, which may be zero or
   * a positive number (not greater than maxOccurrences()).
   */
  abstract public function minOccurrences();

  /**
   * The maximum number of member instances, which may either be
   * a positive number (not less than minOccurrences()) or the
   * TfSchemaCardinality::Unbounded constant.
   */
  abstract public function maxOccurrences();

  /**
   * The member element that is to be repeated.
   */
  abstract public function member();

  /**
   * Return TRUE if no member instances are required, or FALSE if at least one
   * concrete data element must exist that corresponds to this relation.
   */
  public function isOptional() {
    if ($this->minOccurrences() == 0) {
      return TRUE;
    }

    // minOccurrences() is greater than 0, so the result value depends on
    // whether the member itself is optional or not.
    return $this->member()->isOptional();
  }
}

class TfWritableSchemaCardinality extends TfSchemaCardinality {
  private $member;
  private $minOccurrences;
  private $maxOccurrences;

  public function __construct(TfSchemaCollectionMember $member, $minOccurrences, $maxOccurrences) {
    $this->member = $member;
    $this->minOccurrences = $minOccurrences;
    $this->maxOccurrences = $maxOccurrences;
  }

  public function member() {
    return $this->member;
  }
  public function minOccurrences() {
    return $this->minOccurrences;
  }
  public function maxOccurrences() {
    return $this->maxOccurrences;
  }
}

class TfProxySchemaCardinality extends TfSchemaCardinality implements TfSchemaProxy {
  protected $original;
  private $memberProxy;

  final public function __construct(TfSchemaCardinality $original, TfSchemaProxy $parent = NULL) {
    parent::__construct();
    $this->original = $original;
    $this->memberProxy = NULL;
    $this->setParent($parent);
  }
  protected function setParent(TfSchemaProxy $parent) {
    // Default implementation: do nothing.
  }

  public function __clone() {
    if (isset($this->memberProxy)) {
      $this->memberProxy = clone $this->memberProxy;
    }
  }

  public function minOccurrences() {
    return $this->original->minOccurrences();
  }
  public function maxOccurrences() {
    return $this->original->maxOccurrences();
  }

  public function member() {
    if (!isset($this->memberProxy)) {
      $originalMember = $this->original->member();

      if (!isset($originalMember)) {
        $this->memberProxy = NULL;
      }
      else {
        $baseName = $originalMember->proxyBaseName();  // e.g. "SchemaElement"
        $className = $this->proxyClassName($baseName); // e.g. "TfProxySchemaElement"
        $this->memberProxy = new $className($originalMember, $this);
      }
    }
    return $this->memberProxy;
  }
  protected function proxyClassName($baseName) {
    return 'TfProxy' . $baseName;
  }
}


//
// Collections.
//

/**
 * A container for schema elements and/or other nested collections, specifying
 * the allowed order of data elements in a flat list.
 */
abstract class TfSchemaCollection implements TfSchemaPart {
  public function __construct() {
    assert(($this instanceof TfSchemaCompleteKnowledgeCollection)
        || ($this instanceof TfSchemaPartialKnowledgeCollection));
  }

  /**
   * The list of TfSchemaCollectionMember objects that this collection contains.
   */
  abstract public function members();

  /**
   * Create a data-to-schema-associating parser specific to the kind of
   * collection that a derived class represents.
   * (See transformations.schema.parser.inc for more information on that stuff.)
   */
  abstract public function createParser();

  /**
   * Implementation of TfSchemaPart::proxyBaseName().
   */
  final public function proxyBaseName() {
    if ($this instanceof TfSchemaCompleteKnowledgeCollection) {
      return "SchemaCompleteKnowledgeCollection";
    }
    elseif ($this instanceof TfSchemaPartialKnowledgeCollection) {
      return "SchemaPartialKnowledgeCollection";
    }
  }
}

/**
 * Interface for a TfSchemaCollection that describes the full set of elements
 * in the collection that it represents. In other words, a class implementing
 * this interface will only correspond to a list of data elements if those
 * elements exactly match the collection's members and no further elements
 * exist. Lists of data elements will be rejected if any piece of data does not
 * correspond to the collection's members.
 *
 * Most of the widely used schema descriptions (XML Schema, database tables,
 * query results, etc.) can tell exactly which data they contain, and which
 * data they don't contain, and thus map to such a type of collection.
 */
interface TfSchemaCompleteKnowledgeCollection {
}

/**
 * Interface for a TfSchemaCollection that describes only a partial view of the
 * collection that it represents. In other words, a class implementing this
 * interface will correspond to a list of data elements if the elements match
 * the collection's members - but nothing can be told about data that does not
 * correspond to those.
 *
 * An example of such an incomplete view on an object would be RDF data (which
 * is by definition a collection of incomplete knowledge) or a PHP arrays that
 * contains a known set of elements but maybe also other elements that are not
 * known beforehand (and can be ignored).
 *
 * That uncertainty makes this type of collection a difficult target for the
 * parsing process that is being used to correlate data to schema elements,
 * because if elements cannot be rejected for a lack of knowledge, every
 * element must be matched. Because of that property, partial knowledge
 * collections are only allowed as direct child collection of a schema elements
 * where no harm is done by matching all elements.
 */
interface TfSchemaPartialKnowledgeCollection {
}


/**
 * Abstract base class for schema collections that need to manage multiple
 * members in a read/write fashion.
 */
abstract class TfWritableSchemaCollection extends TfSchemaCollection {
  private $members;

  public function __construct() {
    parent::__construct();
    $this->members = array();
  }

  /**
   * An array of TfSchemaCollectionMember objects, with integer values as keys.
   * Depending on the collection type, the order of this list might also
   * indicate the required order of the concrete data elements.
   */
  final public function members() {
    return $this->members;
  }

  /**
   * Retrieve a single member with array key @p $key, or NULL if no member is
   * assigned to that key.
   */
  final public function member($key) {
    return isset($this->members[$key]) ? $this->members[$key] : NULL;
  }

  /**
   * Return TRUE if this collection is empty (does not have any members),
   * or FALSE otherwise.
   */
  final public function isEmpty() {
    return empty($this->members);
  }

  /**
   * Add a member element to this collection.
   *
   * @return
   *   The array key that has been newly assigned to this member element.
   */
  public function addMember(TfSchemaCollectionMember $member) {
    $key = empty($this->members) ? 0 : (max(array_keys($this->members)) + 1);
    $this->members[$key] = $member;
    return $key;
  }

  /**
   * Remove the member with the given @p $key.
   *
   * @return
   *   TRUE if a member was assigned to this key (and removed), or FALSE if
   *   no member was assigned to this key (and no action has been taken as
   *   a consequence).
   */
  public function removeMember($key) {
    if (isset($this->members[$key])) {
      unset($this->members[$key]);
      return TRUE;
    }
    return FALSE;
  }
}

abstract class TfProxySchemaCollection extends TfSchemaCollection implements TfSchemaProxy {
  protected $original;
  private $memberProxies;

  public function __construct(TfSchemaCollection $original, TfSchemaProxy $parent = NULL) {
    parent::__construct();
    $this->original = $original;
    $this->memberProxies = NULL;
    $this->setParent($parent);
  }
  protected function setParent(TfSchemaProxy $parent) {
    // Default implementation: do nothing.
  }

  public function __clone() {
    if (isset($this->memberProxies)) {
      $newMemberProxies = array();
      foreach ($this->memberProxies as $memberProxy) {
        $newMemberProxies[] = clone $memberProxy;
      }
      $this->memberProxies = $newMemberProxies;
    }
  }

  /**
   * Implementation of TfSchemaCollection::createParser().
   */
  public function createParser() {
    $parser = $this->original->createParser();
    $parser->initialize($this);
  }

  /**
   * Implementation of TfSchemaCollection::members().
   */
  public function members() {
    if (!isset($this->memberProxies)) {
      $originalMembers = $this->original->members();
      $this->memberProxies = array();

      foreach ($originalMembers as $originalMember) {
        $baseName = $originalMember->proxyBaseName();  // e.g. "SchemaElement"
        $className = $this->proxyClassName($baseName); // e.g. "TfProxySchemaElement"
        $this->memberProxies = new $className($originalMember, $this);
      }
    }
    return $this->memberProxy;
  }
  protected function proxyClassName($baseName) {
    return 'TfProxy' . $baseName;
  }
}

class TfProxySchemaCompleteKnowledgeCollection extends TfProxySchemaCollection implements TfSchemaCompleteKnowledgeCollection {
  final public function __construct(TfSchemaCompleteKnowledgeCollection $original, TfSchemaProxy $parent = NULL) {
    parent::__construct($original, $parent);
  }
}

class TfProxySchemaPartialKnowledgeCollection extends TfProxySchemaCollection implements TfSchemaPartialKnowledgeCollection {
  final public function __construct(TfSchemaPartialKnowledgeCollection $original, TfSchemaProxy $parent = NULL) {
    parent::__construct($original, $parent);
  }
}


/**
 * A sequence collection: All of the members must exist in the corresponding
 * concrete data, in the exact same order as they appear in the result value of
 * the members() function.
 */
class TfSchemaSequence extends TfWritableSchemaCollection implements TfSchemaCompleteKnowledgeCollection {
  /**
   * Implementation of TfSchemaCollectionMember::isOptional():
   * If all of the members are mandatory then this sequence collection is too.
   */
  public function isOptional() {
    foreach ($this->members() as $key => $member) {
      if (!$member->isOptional()) {
        return FALSE;
      }
    }
    return TRUE;
  }

  /**
   * Implementation of TfSchemaCollection::createParser().
   */
  public function createParser() {
    include_once(drupal_get_path('module', 'transformations') .'/core/transformations.schema.inc');
    return new TfSchemaSequenceParser($this);
  }
}

/**
 * A set collection: All of the members must exist in the corresponding
 * concrete data, in any order. Like for all complete-knowledge collections,
 * additional members are not allowed.
 */
class TfSchemaSet extends TfWritableSchemaCollection implements TfSchemaCompleteKnowledgeCollection {
  /**
   * Implementation of TfSchemaCollectionMember::isOptional():
   * If all of the members are mandatory then this set collection is too.
   */
  public function isOptional() {
    foreach ($this->members() as $key => $member) {
      if (!$member->isOptional()) {
        return FALSE;
      }
    }
    return TRUE;
  }

  /**
   * Implementation of TfSchemaCollection::createParser().
   */
  public function createParser() {
    include_once(drupal_get_path('module', 'transformations') .'/core/transformations.schema.inc');
    return new TfSchemaSetParser($this);
  }
}

/**
 * A partial-knowledge set collection: All of the members must exist in the
 * corresponding concrete data, in any order. Additional data is also allowed
 * in that set.
 */
class TfSchemaSetFragment extends TfWritableSchemaCollection implements TfSchemaPartialKnowledgeCollection {
  /**
   * Implementation of TfSchemaCollection::createParser().
   */
  public function createParser() {
    include_once(drupal_get_path('module', 'transformations') .'/core/transformations.schema.inc');
    $parser = new TfSchemaSetParser($this);
    $parser->setStrict(FALSE);
  }
}

/**
 * A union collection, also known as "choice": Exactly one of the members
 * must exist in the corresponding concrete data.
 */
class TfSchemaUnion extends TfWritableSchemaCollection implements TfSchemaCompleteKnowledgeCollection {
  /**
   * Implementation of TfSchemaCollectionMember::isOptional():
   * If any of the members is optional then this union collection is too.
   */
  public function isOptional() {
    foreach ($this->members() as $key => $member) {
      if ($member->isOptional()) {
        return TRUE;
      }
    }
    return FALSE;
  }

  /**
   * Implementation of TfSchemaCollection::createParser().
   */
  public function createParser() {
    include_once(drupal_get_path('module', 'transformations') .'/core/transformations.schema.inc');
    return new TfSchemaUnionParser($this);
  }
}
