<?php
// $Id: transformations.schema.parser.inc,v 1.1 2009/03/19 19:12:53 jpetso Exp $
/**
 * @file
 * Transformations -
 * A framework for generic schema-based data transformation pipelines.
 *
 * This file contains the parser classes that associate concrete data to
 * schema elements.
 *
 * Copyright 2009 by Jakob Petsovits ("jpetso", http://drupal.org/user/56020)
 */

/**
 * A recursive descent parser, disguised as tool for checking and correlating
 * data to a given schema.
 */
class TfSchemaParser {
  private $parserStack;
  private $currentParser;
  private $encounteredParseError;

  public function __construct(TfSchemaCollection $collection) {
    $this->initialize($collection);
  }

  public function initialize(TfSchemaCollection $collection) {
    $this->parserStack = array();
    $this->currentParser = $collection->createParser();
    $this->encounteredParseError = FALSE;
  }

  private function popParser() {
    // Return to the previous parser in the stack.
    if (empty($this->parserStack)) {
      $this->currentParser = NULL;
    }
    else {
      $this->currentParser = array_pop($this->parserStack);
    }
  }

  private function pushParser($parser) {
    $this->parserStack[] = $this->currentParser;
    $this->currentParser = $parser;
  }

  public function match($label, $value) {
    $tryNext = TRUE;

    while ($tryNext) {
      if ($this->encounteredParseError || !isset($this->currentParser)) {
        return FALSE;
      }
      $match = $this->currentParser->beginMatch($label, $value);

      if ($match->member instanceof TfSchemaCollection)) { // "please recurse"
        $this->pushParser($match->member->createParser());
        return $this->match($label, $value);
      }
      $match = $this->acceptMatch($match);
      $tryNext = !empty($match->tryNext);
    }

    if ($match->successful) {
      $match->data = $value;
    }
    else {
      $this->encounteredParseError = TRUE;
    }
    return $match;
  }

  public function acceptMatch(&$match) {
    $match = $this->currentParser->acceptMatch($match);

    if (!empty($match->notAccepted)) { // go back, up the stack
      unset($match->notAccepted);
      $match->successful = $this->currentParser->wasSuccessful();

      if (isset($match->tryAgain)) {
        unset($match->tryAgain);
      }
      if (!$match->successful) {
        // [backtrack]
        // $match->tryAgain = TRUE;
        return $match;
      }

      $this->popParser();

      if (isset($this->currentParser)) {
        $match = $this->acceptMatch($match);
      }
    }
    return $match;
  }

  /**
   * After all data has been passed to the match() function, this finalizes the
   * process and returns TRUE if all the data matched the schema node or FALSE
   * if any requirement was not satisfied.
   */
  public function wasSuccessful() {
    if ($this->encounteredParseError) {
      return FALSE;
    }
    // Parsers might still be in an "open" state, successful or not.
    // Get all of those off the stack and check if they're successful as is.
    while (isset($this->currentParser)) {
      $match = new stdClass();
      $match->successful = $this->currentParser->wasSuccessful();

      if (!$match->successful) {
        return FALSE;
      }
      $this->popParser();

      if (isset($this->currentParser)) {
        $match = $this->currentParser->acceptMatch($match);
      }
    }
    // All parsers done, no errors encountered: yup, successful
    return TRUE;
  }
}

/**
 * Base class for collection specific parsers.
 */
abstract class TfSchemaCollectionParser {
  protected $currentRelation;

  final public function __construct(TfSchemaCollection $collection) {
    $this->initialize($collection);
  }

  abstract public function initialize(TfSchemaCollection $collection);

  public function beginMatch($label, $value) {
    $match = new stdClass();

    if (!isset($this->currentRelation)) {
      $match->successful = FALSE;
      return $match;
    }

    $match->relation = $this->currentRelation->rel;

    if (!$this->currentRelation->rel->isValidLabel($label)) {
      $match->successful = FALSE;
      return $match;
    }
    $childSchema = $this->currentRelation->rel->childSchema();

    if ($childSchema->isVirtual()) {
      if (!$this->isValidFirstLabel($label, $childSchema)) {
        $match->successful = FALSE;
        return $match;
      }
      $match->virtual = TRUE;
      return $match;
    }

    $match->successful = $this->isValidValue($value); // TODO: implement
    return $match;
  }

  /**
   * Determine if the given label is suitable for the first match of the child
   * relations of the virtual schema node @p $schema, i.e. whether we should
   * descend into @p $schema or not. (In parser terms: this checks if the given
   * label is in the FIRST set of @p $schema.)
   *
   * This check shouldn't be necessary if the schema parser was able to do
   * backtracking, but it's not so this is an additional safety check.
   */
  public function isValidFirstLabel($label, $schema) {
    switch($schema->collectionType()) {
      case TfSchemaElement::CollectionSequence:
        foreach ($schema->childRelations() as $childRelation) {
          if ($childRelation->isValidLabel($label)) {
            break;
          }
          $childSchema = $childRelation->childSchema();
          if ($childSchema->isVirtual() && $this->isValidFirstLabel($label, $childSchema)) {
            break;
          }
          if (!$childRelation->isOptional()) {
            return FALSE;
          }
        }
        // One of the possible first relations accepts this label.
        return TRUE;

      case TfSchemaElement::CollectionSet:
      case TfSchemaElement::CollectionUnion:
        foreach ($schema->childRelations() as $childRelation) {
          $valid = FALSE;
          if ($childRelation->isValidLabel($label)) {
            $valid = TRUE;
          }
          $childSchema = $childRelation->childSchema();

          if ($childSchema->isVirtual()) {
            $valid = FALSE;
            if ($this->isValidFirstLabel($label, $childSchema)) {
              $valid = TRUE;
            }
          }
          if ($valid) {
            return TRUE;
          }
        }
        // None of the child relations accepts this label.
        return FALSE;
    }
    return FALSE;
  }
}

class TfSchemaSequenceParser extends TfSchemaCollectionParser {
  private $remainingRelations;

  public function initialize(TfSchemaCollection $collection) {
    $this->remainingRelations = $schema->childRelations();
    $this->nextRelation();
  }

  private function nextRelation() {
    if (empty($this->remainingRelations)) {
      $this->currentRelation = NULL;
    }
    $relation = array_shift($this->remainingRelations);
    $this->currentRelation = new stdClass();
    $this->currentRelation->rel = $relation;
    $this->currentRelation->occurrences = 0;
  }

  public function acceptMatch(&$match) {
    if (!$match->successful) {
      if ($this->currentRelation->occurrences >= $this->currentRelation->rel->minOccurrences()) {
        // Nevermind the failed match, there are enough occurrences already. So
        // instead of failing, try if the next relation makes for a better fit.
        $this->nextRelation();
        $match->tryNext = TRUE;
        return $match;
      }
      else { // Not enough occurrences.
        $match->notAccepted = TRUE;
        return $match;
      }
    }
    // Ok, so the match is valid. In that case, increment the number of
    // occurrences for the corresponding relation, and proceed to the next
    // relation if we're already at the maximum.
    ++($this->currentRelation->occurrences);

    if ($this->currentRelation->occurrences == $this->currentRelation->rel->maxOccurrences()) {
      $this->nextRelation();
    }
    return $match;
  }

  /**
   * Returns TRUE if the data satisfies the schema node, or FALSE if not.
   */
  public function wasSuccessful() {
    if (isset($this->currentRelation)
        && $this->currentRelation->occurrences < $this->currentRelation->rel->minOccurrences()) {
      return FALSE;
    }
    if (empty($this->remainingRelations)) {
      return TRUE;
    }
    foreach ($this->remainingRelations as $relation) {
      if (!$relation->isOptional()) {
        return FALSE;
      }
    }
    return TRUE;
  }
}


class TfSchemaSetParser extends TfSchemaCollectionParser {
  private $remainingRelations;
  private $currentRelationKey;

  public function initialize(TfSchemaCollection $collection) {
    $this->remainingRelations = array();

    $i = 0;
    $virtualRelations = array();
    foreach ($schema->childRelations() as $rel) {
      $relation = new stdClass();
      $relation->rel = array_shift($rel);
      $relation->occurrences = 0;
      $this->remainingRelations[] = $relation;
      ++$i;
    }
    $this->resetCurrentRelation();
  }

  private function resetCurrentRelation() {
    $this->currentRelationKey = 0;
    $this->currentRelation = empty($this->remainingRelations)
                              ? NULL
                              : $this->remainingRelations[0];
  }

  private function nextRelation() {
    ++($this->currentRelationKey);
    $this->currentRelation = isset($this->remainingRelations[$this->currentRelationKey])
                              ? $this->remainingRelations[$this->currentRelationKey]
                              : NULL;
  }

  public function acceptMatch($match) {
    if (!$match->successful) {
      $this->nextRelation();
      if (isset($this->currentRelation)) {
        // Nevermind the failed match, there are still other options left.
        $match->tryNext = TRUE;
        return $match;
      }
      else { // No options left, fail the match.
        $match->notAccepted = TRUE;
        return $match;
      }
    }
    // Ok, so the match is valid. In that case, increment the number of
    // occurrences for the corresponding relation, and reset the current
    // relation so that we can test all of them again for the next match.
    ++($this->currentRelation->occurrences);

    // If we've reached the maximum number of occurrences, this relation does
    // not interest us anymore. Thus, remove it from the list of options.
    if ($this->currentRelation->occurrences == $this->currentRelation->rel->maxOccurrences()) {
      unset($this->remainingRelations[$this->currentRelationKey]);
      $this->remainingRelations = array_values($this->remainingRelations); // we need continuous indexes
    }

    $this->resetCurrentRelation();
    return $match;
  }

  /**
   * Returns TRUE if the data satisfies the schema node, or FALSE if not.
   */
  public function wasSuccessful() {
    foreach ($this->remainingRelations as $relation) {
      if ($relation->occurrences < $relation->rel->minOccurrences()) {
        return FALSE;
      }
    }
    return TRUE;
  }
}


class TfSchemaUnionParser extends TfSchemaCollectionParser {
  private $remainingRelations;

  public function initialize(TfSchemaCollection $collection) {
    $this->remainingRelations = $schema->childRelations();
    $this->nextRelation();
  }

  private function nextRelation() {
    if (empty($this->remainingRelations)) {
      $this->currentRelation = NULL;
    }
    $relation = array_shift($this->remainingRelations);
    $this->currentRelation = new stdClass();
    $this->currentRelation->rel = $relation;
    $this->currentRelation->occurrences = 0;
  }

  public function acceptMatch(&$match) {
    if (!$match->successful) {
      if (!empty($this->remainingRelations)) {
        // Nevermind the failed match, we've got other options left. Try those.
        $this->nextRelation();
        $match->tryNext = TRUE;
        return $match;
      }
      else { // No options left, fail the match.
        $match->notAccepted = TRUE;
        return $match;
      }
    }
    // Ok, so the match is valid. In that case, increment the number of
    // occurrences for the corresponding relation, and remove all other options
    // as we're now fixed with that single schema element.
    ++($this->currentRelation->occurrences);
    $this->remainingRelations = array();

    if ($this->currentRelation->occurrences == $this->currentRelation->rel->maxOccurrences()) {
      $this->currentRelation = NULL;
    }
    return $match;
  }

  /**
   * Returns TRUE if the data satisfies the schema node, or FALSE if not.
   */
  public function wasSuccessful() {
    if (empty($this->remainingRelations) && !isset($this->currentRelation)) {
      return TRUE;
    }
    if ($this->currentRelation->occurrences >= $this->currentRelation->rel->minOccurrences()) {
      return TRUE;
    }
    return FALSE;
  }
}
