<?php
// $Id: data.xml.inc,v 1.1 2009/03/19 19:12:54 jpetso Exp $
/**
 * @file
 * Transformations -
 * A framework for generic schema-based data transformation pipelines.
 *
 * Copyright 2008 by Jakob Petsovits ("jpetso", http://drupal.org/user/56020)
 */

/**
 * A helper class featuring static methods for DOMNode objects.
 */
class TfXMLTools {
  public static function asXML(DOMNode $node, $formatOutput = TRUE) {
    if ($node->nodeType == XML_DOCUMENT_NODE) {
      $doc = $node;
    }
    else {
      $doc = new DOMDocument();
      $importedNode = $doc->importNode($node, TRUE);
      $doc->appendChild($importedNode);
    }
    $originalFormatOutput = $doc->formatOutput;
    $doc->formatOutput = $formatOutput;
    $xml = $doc->saveXML();
    $doc->formatOutput = $originalFormatOutput;
    return $xml;
  }

  public static function innerXML(DOMNode $node, $formatOutput = TRUE) {
    static $processedChildNodeTypes = array(XML_ELEMENT_NODE, XML_TEXT_NODE);
    $innerXML = '';
    $doc = ($node->nodeType == XML_DOCUMENT_NODE)
      ? $node
      : $node->ownerDocument;

    $originalFormatOutput = $doc->formatOutput;
    $doc->formatOutput = $formatOutput;

    // Loop through all the children, getting the textual representation of
    // each of them.
    $children = $node->childNodes;
    foreach ($children as $child) {
      if (!in_array($child->nodeType, $processedChildNodeTypes)) {
        continue;
      }
      $innerXML .= $doc->saveXML($child);
    }
    $doc->formatOutput = $originalFormatOutput;

    return $innerXML;
  }
}

class TfXMLElementData extends TfConcreteData {
  protected function initialize($data, $importing = FALSE) {
    // We accept plain XML string data for straightforward unserialization.
    if (is_string($data)) {
      $domDocument = new DOMDocument();
      if ($domDocument->loadXML($data, LIBXML_NOCDATA | LIBXML_NOENT | LIBXML_NOBLANKS)) {
        $data = $domDocument;
      }
    }

    // SimpleXML is just not flexible enough to enable iterations like with DOM.
    // Luckily, there's an easy method to convert between those two, so it's
    // possible to have a simple wrapper that takes and delivers SimpleXML data.
    if (is_object($data) && $data instanceof SimpleXMLElement) {
      $domData = dom_import_simplexml($data);
      if ($domData !== FALSE) {
        $data = $domData;
      }
      // otherwise, we keep the original data for the upcoming error message.
    }

    if (!is_object($data) || !($data instanceof DOMNode)) {
      throw new Exception('Tried to initialize a TfXMLElementData object with incompatible data: "' . print_r($data, TRUE) . '" (' . gettype($data) . ')');
    }
    if ($data->nodeType == XML_DOCUMENT_NODE && is_null($data->documentElement)) {
      throw new Exception('An XML document without document element is invalid and thus won\'t be accepted.');
    }
    parent::initialize($data, $importing);
  }

  public function data() {
    if ($this->isLeafElement()) {
      return $this->data->textContent;
    }
    return parent::data();
  }

  protected function serializableData() {
    return TfXMLTools::asXML($this->data);
  }

  public function asDOMNode() {
    return $this->data;
  }

  public function isLeafElement() {
    return !in_array($this->data->nodeType, array(XML_DOCUMENT_NODE, XML_ELEMENT_NODE));
  }

  protected function rawChildElementIterator() {
    $iterator = new AppendIterator();

    if ($this->data->nodeType == XML_ELEMENT_NODE) {
      $baseProperties = array(
        //'#name' => $this->data->localName,
        '#qualifiedName' => self::qualifiedName($this->data),
        //'#prefixedName' => empty($this->data->prefix)
        //                    ? $this->data->localName
        //                    : $this->data->prefix . ':' . $this->data->localName,
        //'#prefix' => empty($this->data->prefix) ? '' : $this->data->prefix,
        //'#namespace' => empty($this->data->namespaceURI) ? '' : $this->data->namespaceURI,
      );
      $iterator->append(new ArrayIterator($baseProperties));

      if (!empty($this->data->attributes)) {
        $attributes = array();
        foreach ($this->data->attributes as $attribute) {
          $attributes['@' . self::qualifiedName($attribute)] = $attribute->value;
        }
        $iterator->append(new ArrayIterator($attributes));
      }
    }
    $iterator->append(new TfDOMElementIterator(new TfDOMNodeListIterator($this->data->childNodes)));
    return $iterator;
  }

  public static function qualifiedName(DOMNode $node) {
    return empty($node->namespaceURI)
      ? $node->localName
      : $node->namespaceURI . ':' . $node->localName;
  }
}

/**
 * Iterator that provides the name of the attribute, prefixed with '@', as key,
 * and its contents as string value.
 *
 * Takes $domNode->attributes as constructor argument.
 */
class TfDOMAttributeIterator extends IteratorIterator {
  public function key() {
    return '@' . parent::current()->name;
  }

  public function current() {
    return parent::current()->value;
  }
}

/**
 * There's no pre-made iterator for DOMNodeList, so let's take the one that has
 * been kindly posted to http://php.net/manual/en/domnodelist.item.php
 * by "vinyanov at poczta dot onet dot pl".
 */
class TfDOMNodeListIterator implements RecursiveIterator {
  private $nodeList;
  private $offset;

  public function __construct(DOMNodeList $nodeList) {
    $this->nodeList = $nodeList;
  }

  public function rewind() {
    $this->offset = 0;
  }

  public function next() {
    $this->offset++;
  }

  public function valid() {
    return ($this->offset < $this->nodeList->length);
  }

  public function current() {
    return $this->nodeList->item($this->offset);
  }

  public function key() {
    return $this->current()->nodeName;
  }

  public function hasChildren() {
    return isset($this->current()->childNodes->length) && $this->current()->childNodes->length > 0;
  }

  public function getChildren() {
    return new self($this->current()->childNodes);
  }
}

/**
 * Iterator that provides the name of the element as key.
 *
 * Takes a TfDOMNodeListIterator as constructor argument, preferably
 * initialized with $domNode->childNodes.
 */
class TfDOMElementIterator extends FilterIterator {
  public function accept() {
    $element = parent::current();
    switch ($element->nodeType) {
      case XML_ELEMENT_NODE:
        return TRUE;
      case XML_TEXT_NODE:
        $text = trim($element->textContent);
        return !empty($text);
      default:
        return FALSE;
    }
  }

  public function key() {
    $element = parent::current();
    if ($element->nodeType == XML_TEXT_NODE) {
      return '#text';
    }
    return empty($element->namespaceURI)
            ? $element->localName
            : $element->namespaceURI . ':' . $element->localName;
  }
}
