<?php
// $Id: operations.foreach.inc,v 1.2 2009/03/24 19:44:59 jpetso Exp $
/**
 * @file
 * Transformations -
 * A framework for generic schema-based data transformation pipelines.
 *
 * Copyright 2008 by Jakob Petsovits ("jpetso", http://drupal.org/user/56020)
 */


/**
 * Implementation of [module]_operation_[class]().
 */
function transformations_operation_TfForEach() {
  return array(
    'label' => t('For-each loop'),
    'description' => t('Executes an operation for each element of a list.'),
    'usesTypeOptions' => TRUE,
  );
}

/**
 * Implementation of [module]_operation_[class]_type_options().
 */
function transformations_operation_TfForEach_type_options(&$typeOptionsInfo, $prefix) {
  $operations = transformations_operation_info();

  $typeOptionsInfo[$prefix . 'innerOperation']['label'] = t('Inner operation');
  $typeOptionsInfo[$prefix . 'innerOperation']['description'] =
    t('The operation that will be executed once for each value in the list.');
  $typeOptionsInfo[$prefix . 'innerOperation']['options'] = array();

  // Add each possible operation as inner operation of TfForEach.
  if (!isset($typeOptionsInfo[$prefix . 'innerOperation']['value'])) {
    foreach ($operations as $className => $info) {
      $typeOptionsInfo[$prefix . 'innerOperation']['options'][$className] = $info['label'];
    }
  }
  else {
    $className = $typeOptionsInfo[$prefix . 'innerOperation']['value'];
    $typeOptionsInfo[$prefix . 'innerOperation']['options'][$className] =
      $operations[$className]['label'];

    $status = TfOperationTools::processTypeOptionsInfo(
      $className, $typeOptionsInfo, $prefix . 'inner_'
    );

    if ($status == TfOperationTools::TypeOptionsDone) {
      $typeOptions = TfOperationTools::typeOptionsValuesFromInfo($typeOptionsInfo);
      $typeOptions = TfOperation::typeOptionsRemovePrefix($typeOptions, $prefix . 'inner_');
      $operation = TfOperation::load($className, $typeOptions);

      $typeOptionsInfo[$prefix . 'listInput']['label'] = t('List input');
      $typeOptionsInfo[$prefix . 'listInput']['description'] =
        t('The inner operation\'s input that will be passed each list element separately.');

      if ($operation) {
        $inputKeys = $operation->inputKeys();
        foreach ($inputKeys as $inputKey) {
          $typeOptionsInfo[$prefix . 'listInput']['options'][$inputKey] =
            check_plain($operation->inputProperty($inputKey, 'label'));
        }
      }
      $typeOptionsInfo[$prefix . 'listInput']['options']
        [TfForEach::ListInputNone] = t('(none)');
    }
  }
}


/**
 * An operation that takes a list of data elements and executes the given
 * inner operation once for each of those elements. Possible other inputs of
 * the inner operation are passed from the foreach operation without being
 * unfolded from any list.
 */
class TfForEach extends TfOperation implements TfOutputSchemaListener {
  const ListInputNone = 0;

  /** The instance of the operation that is executed for each element.
   * NULL if the operation has not yet been instantiated. */
  private $innerOperation;

  /** The input key for the input of the inner operation that will be fed
   * the separate elements passed as a list to this foreach operation. */
  private $listInput;

  /** The input key that this foreach operation uses for the list input.
   * This is usually $listInput . 'List', except if the inner operation
   * already has an input that is named this way, in which case
   * $outerListInput is the same as $listInput. */
  private $outerListInput;


  /**
   * Implementation of TfOperation::initialize().
   */
  public function initialize($typeOptions = array()) {
    if (isset($this->innerOperation) && is_object($this->innerOperation)) {
      $this->innerOperation->unregisterOutputSchemaListener($this);
    }

    parent::initialize($typeOptions);
    $this->innerOperation = FALSE;
    $this->listInput = $typeOptions['listInput'];

    // Assign the inner operation, newly instantiating it if necessary.
    if (is_string($typeOptions['innerOperation'])) {
      $innerTypeOptions = TfOperation::typeOptionsRemovePrefix($typeOptions, 'inner_');
      $this->innerOperation = TfOperation::load($typeOptions['innerOperation'], $innerTypeOptions);
    }
    elseif (is_object($typeOptions['innerOperation']) && $typeOptions['innerOperation'] instanceof TfOperation) {
      $this->innerOperation = $typeOptions['innerOperation'];
    }

    if (!is_object($this->innerOperation)) {
      throw new Exception('Operation "' .
        check_plain($typeOptions['innerOperation']) . '" cannot be loaded!');
    }

    if (!is_string($this->listInput) && !is_numeric($this->listInput)) {
      throw new Exception('Type option "listInput" must be a valid input key type (string).');
    }

    // Register ourselves with any output schema change notifications.
    $this->innerOperation->registerOutputSchemaListener($this);

    // Make sure that the list input key actually exists.
    $inputKeys = $this->innerOperation->inputKeys();

    // If listInput is FALSE, no input is being assigned the list value,
    // instead, the operation just executes normally the number of times that
    // the input list contains (without processing its actual values).
    if ($this->listInput === TfForEach::ListInputNone) {
      $this->outerListInput = 'foreachList';

      while (in_array($this->outerListInput, $inputKeys)) { // make sure to avoid input key clashes
        preg_match('/foreachList([\d]*)/', $this->outerListInput, $matches);
        $counter = empty($matches[1]) ? 0 : $matches[1];
        $this->outerListInput = 'foreachList' . $counter;
      }
    }
    else {
      if (!in_array($this->listInput, $inputKeys)) {
        throw new Exception('No input "' .
          check_plain($this->listInput) . '" exists for operation "' .
          check_plain(get_class($this->innerOperation)) . '"!');
      }

      $this->outerListInput = $this->listInput . 'List';
      if (in_array($this->outerListInput, $inputKeys)) { // make sure to avoid input key clashes
        $this->outerListInput = $this->listInput;
      }
    }
  }

  /**
   * Overriding TfOperation::hasSideEffects().
   */
  public function hasSideEffects() {
    return $this->innerOperation->hasSideEffects();
  }

  /**
   * Overriding TfOperation::typeOptions().
   */
  public function typeOptions() {
    $typeOptions = array(
      'innerOperation' => get_class($this->innerOperation),
      'listInput' => $this->listInput,
    );
    foreach ($this->innerOperation->typeOptions() as $key => $value) {
      $typeOptions['inner_' . $key] = $value;
    }
    return $typeOptions;
  }

  /**
   * Implementation of TfOperation::inputs().
   */
  protected function inputs() {
    $inputKeys = $this->innerOperation->inputKeys();

    if ($this->listInput !== TfForEach::ListInputNone) {
      $index = array_search($this->listInput, $inputKeys);
      if ($index !== FALSE) {
        $inputKeys[$index] = $this->outerListInput;
      }
    }
    return $inputKeys;
  }

  /**
   * Implementation of TfOperation::inputInfo().
   */
  protected function inputInfo($inputKey, $propertyKey) {
    // All inputs except for the list input are simply relayed with no change.
    if ($inputKey != $this->outerListInput) {
      return $this->innerOperation->inputProperty($inputKey, $propertyKey);
    }

    // The list input is practically a whole new slot, and needs special treatment.
    switch ($propertyKey) {
      case 'required':
        return FALSE;

      case 'label':
        $label = $this->innerOperation->inputProperty($this->listInput, 'label');
        return t('!element list', array('!element' => $label));

      case 'expectedSchema':
        $originalElement = ($this->listInput === TfForEach::ListInputNone)
          ? new TfWritableSchemaElement('data') // takes any element, we don't care
          : $this->innerOperation->inputProperty($this->listInput, 'expectedSchema');

        // Put the original element in a list of original elements.
        $cardinality = new TfWritableSchemaCardinality($originalElement,
          0, TfSchemaCardinality::Unbounded
        );
        $collection = new TfSchemaSet();
        $collection->addMember($cardinality);

        $listElement = new TfWritableSchemaElement($this->outerListInput);
        $listElement->setChildCollection($collection);
        return $listElement;
    }
  }

  /**
   * Implementation of TfOperation::outputs().
   */
  protected function outputs() {
    return array('outputElements');
  }

  /**
   * Implementation of TfOperation::outputInfo().
   */
  protected function outputInfo($outputKey, $propertyKey) {
    if ($outputKey == 'outputElements') {
      switch ($propertyKey) {
        case 'label':
          return t('List of outputs');

        case 'expectedSchema':
          // Wrap the original outputs in a structure collection (which is the
          // original output schema), and the set of all structures in
          // another list (which is the foreach operation's output schema).
          $innerOutputKeys = $this->innerOperation->outputKeys();
          $structureCollection = new TfSchemaSequence();

          foreach ($innerOutputKeys as $innerKey) {
            $innerSchema = $this->innerOperation->outputProperty($innerKey, 'expectedSchema');
            $structureCollection->addMember($innerSchema);
          }
          $structureElement = new TfWritableSchemaElement('output');
          $structureElement->setChildCollection($structureCollection);

          $cardinality = new TfWritableSchemaCardinality($structureElement,
            0, TfSchemaCardinality::Unbounded
          );
          $listCollection = new TfSchemaSequence();
          $listCollection->addMember($cardinality);

          $listElement = new TfWritableSchemaElement('outputElements');
          $listElement->setChildCollection($listCollection);
          return $listElement;
      }
    }
  }

  /**
   * Overriding TfOperation::inputSchemaChanged() in order to relay the new
   * schema to the inner operation and at the same time narrow down output
   * schema change notifications (which are handled by outputSchemaChanged()).
   */
  protected function inputSchemaChanged($inputKey) {
    $schema = $this->inputProperty($inputKey, 'assignedSchema');

    if ($inputKey == $this->outerListInput) {
      if ($this->listInput == TfForEach::ListInputNone) {
        // The unmapped (dead-end) list input was changed - nothing to relay.
        return;
      }
      // The outer operation gets a list element set, so we need to extract its
      // child element(s) that should be passed to the inner operation.
      $inputKey = $this->listInput;
      ///TODO (but not all too important for now)
      // $schema = (...)
      // Until we've got that element extracted, let's ignore this event and
      // just do the bare minimum: calling the default implementation.
      parent::inputSchemaChanged($inputKey);
      return;
    }
    $this->innerOperation->setInputSchema($inputKey, $schema);
  }

  /**
   * Implementation of TfOutputSchemaListener::outputSchemaChanged().
   */
  public function outputSchemaChanged(TfOperation $changedOperation, $outputKey = NULL) {
    // $changedOperation must always be $this->innerOperation, because
    // we don't register ourselves with any other operation.
    // So no need for object comparisons.
    $this->setOutputSchemaDirty($outputKey);
  }

  /**
   * Called when input data has been (re-)set using setInput().
   */
  protected function inputChanged($inputKey) {
    parent::inputChanged($inputKey);

    // Relay all inputs except for the list input which is being
    // handled separately in execute().
    if ($inputKey != $this->outerListInput) {
      $this->innerOperation->setInput($inputKey, $this->input($inputKey));
    }
  }

  /**
   * Implementation of TfOperation::execute().
   * This is just a bit fake, as we defer execution until the actual results
   * are requested from the data wrapper itself. Haha.
   */
  protected function execute(TfOutput $output) {
    $data = (object) array(
      'innerOperation' => clone $this->innerOperation, // clone == fixate inputs
      'listInputKey' => $this->listInput,
      'listInputData' => $this->input($this->outerListInput),
    );
    $wrapper = TfDataWrapper::create(new TfForEachIterator($data));
    $output->set('outputElements', $wrapper);
  }
}

class TfForEachIterator extends IteratorIterator {
  private $innerOperation;
  private $listInputKey;

  public function __construct($data) {
    $this->innerOperation = $data->innerOperation;
    $this->listInputKey = $data->listInputKey;

    parent::__construct($data->listInputData->children(TfDataWrapper::ChildrenMixed));
  }

  public function key() {
    return 'output';
  }

  public function current() {
    if ($this->listInputKey !== TfForEach::ListInputNone) {
      $currentListElement = parent::current();
      $this->innerOperation->setInput($this->listInputKey, $currentListElement);
    }
    return $this->innerOperation->output();

    // ...you may ask, why can we return a TfOutput object as value?
    // It goes like this: TfOutput is an object, so when being wrapped,
    // it will end up in a TfObjectData wrapper. When rawChildElementIterator()
    // is called, TfObjectData returns the TfOutput object itself because it
    // is an iterator. The combination of all of that enables tree navigation
    // by just returning a traversable object. Cool, eh?
  }
}
