<?php
// $Id: operations.xml.inc,v 1.1 2009/03/19 19:12:54 jpetso Exp $
/**
 * @file
 * Transformations -
 * A framework for generic schema-based data transformation pipelines.
 *
 * Copyright 2009 by Jakob Petsovits ("jpetso", http://drupal.org/user/56020)
 */


/**
 * Implementation of [module]_operation_[class]().
 */
function transformations_operation_TfXMLTreeFromXMLString() {
  return array(
    'label' => t('XML string to XML element (tree structure)'),
    'description' => t('Transforms a valid XML string into a tree structure, enabling easier access of the XML elements.'),
  );
}

class TfXMLTreeFromXMLString extends TfOperation {
  /**
   * Implementation of TfOperation::inputs().
   */
  protected function inputs() {
    return array('xmlString');
  }

  /**
   * Implementation of TfOperation::inputInfo().
   */
  protected function inputInfo($inputKey, $propertyKey) {
    if ($inputKey == 'xmlString') {
      switch ($propertyKey) {
        case 'label':
          return t('XML string');

        case 'expectedSchema':
          return new TfWritableSchemaElement('xmlString', 'php:type:string');
      }
    }
  }

  /**
   * Implementation of TfOperation::outputs().
   */
  protected function outputs() {
    return array('xmlTree');
  }

  /**
   * Implementation of TfOperation::outputInfo().
   */
  protected function outputInfo($outputKey, $propertyKey) {
    if ($outputKey == 'xmlTree') {
      switch ($propertyKey) {
        case 'label':
          return t('XML tree');

        case 'expectedSchema':
          return new TfWritableSchemaElement('xmlTree', 'transformations:xml:element', 'php:type:object', 'php:class:DOMDocument');
      }
    }
  }

  /**
   * Implementation of TfOperation::execute().
   */
  protected function execute(TfOutput $output) {
    try {
      TfDataWrapper::includeWrapper('TfXMLElementData');
      $xmlWrapper = new TfXMLElementData($this->input('xmlString')->data());
    }
    catch (Exception $e) {
      $output->setErrorMessage(t('Unable to parse XML string: !message', array(
        '!message' => $e->getMessage(),
      )));
      return;
    }
    $output->set('xmlTree', $xmlWrapper);
  }
}


/**
 * Implementation of [module]_operation_[class]().
 */
function transformations_operation_TfXMLTreeFromXMLFile() {
  return array(
    'label' => t('XML file to XML element (tree structure)'),
    'description' => t('Transforms a valid XML file into a tree structure, enabling easier access of the XML elements.'),
  );
}

class TfXMLTreeFromXMLFile extends TfOperation {
  /**
   * Implementation of TfOperation::inputs().
   */
  protected function inputs() {
    return array('filepath');
  }

  /**
   * Implementation of TfOperation::inputInfo().
   */
  protected function inputInfo($inputKey, $propertyKey) {
    if ($inputKey == 'filepath') {
      switch ($propertyKey) {
        case 'label':
          return t('File path');

        case 'expectedSchema':
          return new TfWritableSchemaElement('filepath', 'php:type:string');
      }
    }
  }

  /**
   * Implementation of TfOperation::outputs().
   */
  protected function outputs() {
    return array('xmlTree');
  }

  /**
   * Implementation of TfOperation::outputInfo().
   */
  protected function outputInfo($outputKey, $propertyKey) {
    if ($outputKey == 'xmlTree') {
      switch ($propertyKey) {
        case 'label':
          return t('XML tree');

        case 'expectedSchema':
          return new TfWritableSchemaElement('xmlTree', 'transformations:xml:element', 'php:type:object', 'php:class:DOMDocument');
      }
    }
  }

  /**
   * Implementation of TfOperation::execute().
   */
  protected function execute(TfOutput $output) {
    $doc = new DOMDocument();
    $filename = realpath(trim($this->input('filepath')->data()));
    if ($doc->load($filename, LIBXML_NOCDATA | LIBXML_NOENT | LIBXML_NOBLANKS)) {
      $output->set('xmlTree', $doc);
    }
    else {
      $output->setErrorMessage(t('Unable to load XML file !filename.', array(
        '!filename' => $filename,
      )));
    }
  }
}


/**
 * Implementation of [module]_operation_[class]().
 */
function transformations_operation_TfOuterXMLStringFromXMLTree() {
  return array(
    'label' => t('XML element (tree structure) to XML string'),
    'description' => t('Transforms an XML element (tree structure) into a valid XML string, including the XML header.'),
  );
}

class TfOuterXMLStringFromXMLTree extends TfOperation {
  /**
   * Implementation of TfOperation::inputs().
   */
  protected function inputs() {
    return array('xmlTree');
  }

  /**
   * Implementation of TfOperation::inputInfo().
   */
  protected function inputInfo($inputKey, $propertyKey) {
    if ($inputKey == 'xmlTree') {
      switch ($propertyKey) {
        case 'label':
          return t('XML tree');

        case 'expectedSchema':
          return new TfWritableSchemaElement('xmlTree', 'transformations:xml:element');
      }
    }
  }

  /**
   * Implementation of TfOperation::outputs().
   */
  protected function outputs() {
    return array('xmlString');
  }

  /**
   * Implementation of TfOperation::outputInfo().
   */
  protected function outputInfo($outputKey, $propertyKey) {
    if ($outputKey == 'xmlString') {
      switch ($propertyKey) {
        case 'label':
          return t('XML string');

        case 'expectedSchema':
          return new TfWritableSchemaElement('xmlString', 'php:type:string');
      }
    }
  }

  /**
   * Implementation of TfOperation::execute().
   */
  protected function execute(TfOutput $output) {
    TfDataWrapper::includeWrapper('TfXMLElementData'); // for TfXMLTools
    $node = $this->input('xmlTree')->data();
    $output->set('xmlString', TfXMLTools::asXML($node));
  }
}


/**
 * Implementation of [module]_operation_[class]().
 */
function transformations_operation_TfInnerXMLStringFromXMLTree() {
  return array(
    'label' => t('XML element (tree structure) to inner XML string'),
    'description' => t('Extracts the contents of an XML element (given as XML tree structure) into a simple string. You probably want to use this in order to make '),
  );
}

class TfInnerXMLStringFromXMLTree extends TfOperation {
  /**
   * Implementation of TfOperation::inputs().
   */
  protected function inputs() {
    return array('xmlTree');
  }

  /**
   * Implementation of TfOperation::inputInfo().
   */
  protected function inputInfo($inputKey, $propertyKey) {
    if ($inputKey == 'xmlTree') {
      switch ($propertyKey) {
        case 'label':
          return t('XML tree');

        case 'expectedSchema':
          return new TfWritableSchemaElement('xmlTree', 'transformations:xml:element');
      }
    }
  }

  /**
   * Implementation of TfOperation::outputs().
   */
  protected function outputs() {
    return array('stringContents');
  }

  /**
   * Implementation of TfOperation::outputInfo().
   */
  protected function outputInfo($outputKey, $propertyKey) {
    if ($outputKey == 'stringContents') {
      switch ($propertyKey) {
        case 'label':
          return t('String contents');

        case 'expectedSchema':
          return new TfWritableSchemaElement('stringContents', 'php:type:string');
      }
    }
  }

  /**
   * Implementation of TfOperation::execute().
   */
  protected function execute(TfOutput $output) {
    TfDataWrapper::includeWrapper('TfXMLElementData'); // for TfXMLTools
    $node = $this->input('xmlTree')->data();
    $output->set('stringContents', TfXMLTools::innerXML($node));
  }
}
